# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division

import copy

from mapproxy.grid import NoTiles, _create_tile_list
from mapproxy.grid.meta_grid import MetaGrid
from mapproxy.grid.resolutions import (
    resolutions,
    ResolutionRange,
    merge_resolution_range,
    resolution_range,
)
from mapproxy.grid.tile_grid import (
    tile_grid,
    TileGrid,
    tile_grid_from_ogc_tile_matrix_set,
    UnsupportedException,
)
from mapproxy.srs import SRS
from mapproxy.util.bbox import TransformationError, bbox_intersects, bbox_contains

import pytest


class TestResolution(object):
    def test_min_res(self):
        conf = dict(min_res=1000)
        res = resolutions(**conf)
        assert res[:5] == [1000, 500.0, 250.0, 125.0, 62.5]
        assert len(res) == 20

    def test_min_res_max_res(self):
        conf = dict(min_res=1000, max_res=80)
        res = resolutions(**conf)
        assert res == [1000, 500.0, 250.0, 125.0]

    def test_min_res_levels(self):
        conf = dict(min_res=1600, num_levels=5)
        res = resolutions(**conf)
        assert res == [1600, 800.0, 400.0, 200.0, 100.0]

    def test_min_res_levels_res_factor(self):
        conf = dict(min_res=1600, num_levels=4, res_factor=4.0)
        res = resolutions(**conf)
        assert res == [1600, 400.0, 100.0, 25.0]

    def test_min_res_levels_sqrt2(self):
        conf = dict(min_res=1600, num_levels=5, res_factor="sqrt2")
        res = resolutions(**conf)
        assert res == pytest.approx([1600.0, 1131.0, 800.0, 566.0, 400.0], 0.1)

    def test_min_res_max_res_levels(self):
        conf = dict(min_res=1600, max_res=10, num_levels=10)
        res = resolutions(**conf)
        assert len(res) == 10
        # will calculate log10 based factor of 1.75752...
        assert res[0] == pytest.approx(1600)
        assert res[1] == pytest.approx(1600 / 1.75752, 0.01)
        assert res[8] == pytest.approx(1600 / 1.75752**8, 0.01)
        assert res[9] == pytest.approx(10)

    def test_bbox_levels(self):
        conf = dict(bbox=[0, 40, 15, 50], num_levels=10, tile_size=(256, 256))
        res = resolutions(**conf)
        assert len(res) == 10
        assert res[0] == pytest.approx(15 / 256)
        assert res[1] == pytest.approx(15 / 512)


class TestAlignedGrid(object):
    def test_epsg_4326_bbox(self):
        base = tile_grid(srs="epsg:4326")
        bbox = (10.0, -20.0, 40.0, 10.0)
        sub = tile_grid(align_with=base, bbox=bbox)

        assert sub.bbox == bbox
        assert sub.resolution(0) == 180 / 256 / 8
        abbox, grid_size, tiles = sub.get_affected_level_tiles(bbox, 0)
        assert abbox == (10.0, -20.0, 55.0, 25.0)
        assert grid_size == (2, 2)
        assert list(tiles) == [(0, 1, 0), (1, 1, 0), (0, 0, 0), (1, 0, 0)]

    def test_epsg_4326_bbox_from_sqrt2(self):
        base = tile_grid(srs="epsg:4326", res_factor="sqrt2")
        bbox = (10.0, -20.0, 40.0, 10.0)
        sub = tile_grid(align_with=base, bbox=bbox, res_factor=2.0)

        assert sub.bbox == bbox
        assert sub.resolution(0) == base.resolution(8)
        assert sub.resolution(1) == base.resolution(10)
        assert sub.resolution(2) == base.resolution(12)

    def test_epsg_4326_bbox_to_sqrt2(self):
        base = tile_grid(srs="epsg:4326", res_factor=2.0)
        bbox = (10.0, -20.0, 40.0, 10.0)
        sub = tile_grid(align_with=base, bbox=bbox, res_factor="sqrt2")

        assert sub.bbox == bbox
        assert sub.resolution(0) == base.resolution(4)
        assert sub.resolution(2) == base.resolution(5)
        assert sub.resolution(4) == base.resolution(6)

        assert sub.resolution(0) > sub.resolution(1) > sub.resolution(3)
        assert sub.resolution(3) / 2 == sub.resolution(5)


def test_metagrid_tiles():
    mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
    assert list(mgrid.meta_tile((0, 0, 0)).tile_patterns) == [((0, 0, 0), (0, 0))]
    assert list(mgrid.meta_tile((0, 1, 1)).tile_patterns) == [
        ((0, 1, 1), (0, 0)),
        ((1, 1, 1), (256, 0)),
        ((0, 0, 1), (0, 256)),
        ((1, 0, 1), (256, 256)),
    ]

    assert list(mgrid.meta_tile((1, 2, 2)).tile_patterns) == [
        ((0, 3, 2), (0, 0)),
        ((1, 3, 2), (256, 0)),
        ((0, 2, 2), (0, 256)),
        ((1, 2, 2), (256, 256)),
    ]


def test_metagrid_tiles_w_meta_size():
    mgrid = MetaGrid(grid=TileGrid(), meta_size=(4, 2))
    assert list(mgrid.meta_tile((1, 2, 2)).tile_patterns) == [
        ((0, 3, 2), (0, 0)),
        ((1, 3, 2), (256, 0)),
        ((2, 3, 2), (512, 0)),
        ((3, 3, 2), (768, 0)),
        ((0, 2, 2), (0, 256)),
        ((1, 2, 2), (256, 256)),
        ((2, 2, 2), (512, 256)),
        ((3, 2, 2), (768, 256)),
    ]


class TestMetaGridGeodetic(object):
    def setup_method(self):
        self.mgrid = MetaGrid(
            grid=tile_grid("EPSG:4326"), meta_size=(2, 2), meta_buffer=10
        )

    def test_meta_bbox_level_0(self):
        assert self.mgrid._meta_bbox((0, 0, 0)) == (
            (-180, -90, 180, 90),
            (0, 0, 0, -128),
        )
        assert self.mgrid._meta_bbox((0, 0, 0), limit_to_bbox=False) == (
            (-194.0625, -104.0625, 194.0625, 284.0625),
            (10, 10, 10, 10),
        )

        assert self.mgrid.meta_tile((0, 0, 0)).size == (256, 128)

    def test_tiles_level_0(self):
        meta_tile = self.mgrid.meta_tile((0, 0, 0))
        assert meta_tile.size == (256, 128)
        assert meta_tile.grid_size == (1, 1)
        assert meta_tile.tile_patterns == [((0, 0, 0), (0, -128))]

    def test_meta_bbox_level_1(self):
        assert self.mgrid._meta_bbox((0, 0, 1)) == ((-180, -90, 180, 90), (0, 0, 0, 0))
        assert self.mgrid._meta_bbox((0, 0, 1), limit_to_bbox=False) == (
            (-187.03125, -97.03125, 187.03125, 97.03125),
            (10, 10, 10, 10),
        )
        assert self.mgrid.meta_tile((0, 0, 1)).size == (512, 256)

    def test_tiles_level_1(self):
        assert list(self.mgrid.meta_tile((0, 0, 1)).tile_patterns) == [
            ((0, 0, 1), (0, 0)),
            ((1, 0, 1), (256, 0)),
        ]

    def test_tile_list_level_1(self):
        assert list(self.mgrid.tile_list((0, 0, 1))) == [(0, 0, 1), (1, 0, 1)]

    def test_meta_bbox_level_2(self):
        assert self.mgrid._meta_bbox((0, 0, 2)) == (
            (-180, -90, 3.515625, 90),
            (0, 0, 10, 0),
        )
        assert self.mgrid._meta_bbox((0, 0, 2), limit_to_bbox=False) == (
            (-183.515625, -93.515625, 3.515625, 93.515625),
            (10, 10, 10, 10),
        )
        assert self.mgrid.meta_tile((0, 0, 2)).size == (522, 512)

        assert self.mgrid._meta_bbox((2, 0, 2)) == (
            (-3.515625, -90, 180, 90),
            (10, 0, 0, 0),
        )
        meta_tile = self.mgrid.meta_tile((2, 0, 2))
        assert meta_tile.size == (522, 512)
        assert meta_tile.grid_size == (2, 2)

    def test_tiles_level_2(self):
        assert list(self.mgrid.meta_tile((0, 0, 2)).tile_patterns) == [
            ((0, 1, 2), (0, 0)),
            ((1, 1, 2), (256, 0)),
            ((0, 0, 2), (0, 256)),
            ((1, 0, 2), (256, 256)),
        ]
        assert list(self.mgrid.meta_tile((2, 0, 2)).tile_patterns) == [
            ((2, 1, 2), (10, 0)),
            ((3, 1, 2), (266, 0)),
            ((2, 0, 2), (10, 256)),
            ((3, 0, 2), (266, 256)),
        ]

    def test_tile_list_level_2(self):
        assert list(self.mgrid.tile_list((0, 0, 2))) == [
            (0, 1, 2),
            (1, 1, 2),
            (0, 0, 2),
            (1, 0, 2),
        ]
        assert list(self.mgrid.tile_list((1, 1, 2))) == [
            (0, 1, 2),
            (1, 1, 2),
            (0, 0, 2),
            (1, 0, 2),
        ]

    def test_tiles_level_3(self):
        assert list(self.mgrid.meta_tile((2, 0, 3)).tile_patterns) == [
            ((2, 1, 3), (10, 10)),
            ((3, 1, 3), (266, 10)),
            ((2, 0, 3), (10, 266)),
            ((3, 0, 3), (266, 266)),
        ]
        assert list(self.mgrid.meta_tile((2, 2, 3)).tile_patterns) == [
            ((2, 3, 3), (10, 0)),
            ((3, 3, 3), (266, 0)),
            ((2, 2, 3), (10, 256)),
            ((3, 2, 3), (266, 256)),
        ]


class TestMetaGridGeodeticUL(object):
    def setup_method(self):
        self.tile_grid = tile_grid("EPSG:4326", origin="ul")
        self.mgrid = MetaGrid(grid=self.tile_grid, meta_size=(2, 2), meta_buffer=10)

    def test_tiles_level_0(self):
        meta_tile = self.mgrid.meta_tile((0, 0, 0))
        assert meta_tile.bbox == pytest.approx((-180, -90, 180, 90))
        assert meta_tile.size == (256, 128)
        assert meta_tile.grid_size == (1, 1)
        assert meta_tile.tile_patterns == [((0, 0, 0), (0, 0))]

    def test_tiles_level_1(self):
        meta_tile = self.mgrid.meta_tile((0, 0, 1))
        assert meta_tile.bbox == pytest.approx((-180, -90, 180, 90))
        assert meta_tile.size == (512, 256)
        assert meta_tile.grid_size == (2, 1)
        assert list(meta_tile.tile_patterns) == [
            ((0, 0, 1), (0, 0)),
            ((1, 0, 1), (256, 0)),
        ]

    def test_tile_list_level_1(self):
        assert list(self.mgrid.tile_list((0, 0, 1))) == [(0, 0, 1), (1, 0, 1)]

    def test_tiles_level_2(self):
        meta_tile = self.mgrid.meta_tile((0, 0, 2))
        assert meta_tile.bbox == pytest.approx((-180, -90, 3.515625, 90))
        assert meta_tile.size == (522, 512)
        assert meta_tile.grid_size == (2, 2)
        assert meta_tile.tile_patterns == [
            ((0, 0, 2), (0, 0)),
            ((1, 0, 2), (256, 0)),
            ((0, 1, 2), (0, 256)),
            ((1, 1, 2), (256, 256)),
        ]
        assert list(self.mgrid.meta_tile((2, 0, 2)).tile_patterns) == [
            ((2, 0, 2), (10, 0)),
            ((3, 0, 2), (266, 0)),
            ((2, 1, 2), (10, 256)),
            ((3, 1, 2), (266, 256)),
        ]

    def test_tile_list_level_2(self):
        assert list(self.mgrid.tile_list((0, 0, 2))) == [
            (0, 0, 2),
            (1, 0, 2),
            (0, 1, 2),
            (1, 1, 2),
        ]
        assert list(self.mgrid.tile_list((1, 1, 2))) == [
            (0, 0, 2),
            (1, 0, 2),
            (0, 1, 2),
            (1, 1, 2),
        ]

    def test_tiles_level_3(self):
        meta_tile = self.mgrid.meta_tile((2, 0, 3))
        assert meta_tile.bbox == pytest.approx((-91.7578125, -1.7578125, 1.7578125, 90))
        assert meta_tile.size == (532, 522)
        assert meta_tile.grid_size == (2, 2)
        assert list(self.mgrid.meta_tile((2, 0, 3)).tile_patterns) == [
            ((2, 0, 3), (10, 0)),
            ((3, 0, 3), (266, 0)),
            ((2, 1, 3), (10, 256)),
            ((3, 1, 3), (266, 256)),
        ]
        assert list(self.mgrid.meta_tile((2, 2, 3)).tile_patterns) == [
            ((2, 2, 3), (10, 10)),
            ((3, 2, 3), (266, 10)),
            ((2, 3, 3), (10, 266)),
            ((3, 3, 3), (266, 266)),
        ]


class TestMetaTile(object):
    def setup_method(self):
        self.mgrid = MetaGrid(
            grid=tile_grid("EPSG:4326"), meta_size=(2, 2), meta_buffer=10
        )

    def test_meta_tile(self):
        meta_tile = self.mgrid.meta_tile((2, 0, 2))
        assert meta_tile.size == (522, 512)

    def test_metatile_bbox(self):
        mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        meta_tile = mgrid.meta_tile((0, 0, 2))
        assert meta_tile.bbox == pytest.approx(
            (-20037508.342789244, -20037508.342789244, 0.0, 0.0)
        )
        meta_tile = mgrid.meta_tile((1, 1, 2))
        assert meta_tile.bbox == pytest.approx(
            (-20037508.342789244, -20037508.342789244, 0.0, 0.0)
        )
        meta_tile = mgrid.meta_tile((4, 5, 3))
        assert meta_tile.bbox == pytest.approx(
            (0.0, 0.0, 10018754.171394622, 10018754.171394622)
        )

    def test_metatile_non_default_meta_size(self):
        mgrid = MetaGrid(grid=TileGrid(), meta_size=(4, 2))
        meta_tile = mgrid.meta_tile((4, 5, 3))
        assert meta_tile.bbox == pytest.approx(
            (0.0, 0.0, 20037508.342789244, 10018754.171394622)
        )
        assert meta_tile.size == (1024, 512)
        assert meta_tile.grid_size == (4, 2)


class TestMetaTileSQRT2(object):
    def setup_method(self):
        self.grid = tile_grid("EPSG:4326", res_factor="sqrt2")
        self.mgrid = MetaGrid(grid=self.grid, meta_size=(4, 4), meta_buffer=10)

    def test_meta_tile(self):
        meta_tile = self.mgrid.meta_tile((0, 0, 8))
        assert meta_tile.size == (1034, 1034)

    def test_metatile_bbox(self):
        meta_tile = self.mgrid.meta_tile((0, 0, 2))
        assert meta_tile.bbox == pytest.approx((-180, -90, 180, 90))
        assert meta_tile.size == (512, 256)
        assert meta_tile.grid_size == (2, 1)
        assert meta_tile.tile_patterns == [((0, 0, 2), (0, 0)), ((1, 0, 2), (256, 0))]

        meta_tile = self.mgrid.meta_tile((1, 0, 2))
        assert meta_tile.bbox == pytest.approx((-180.0, -90, 180.0, 90.0))
        assert meta_tile.size == (512, 256)
        assert meta_tile.grid_size == (2, 1)

        meta_tile = self.mgrid.meta_tile((0, 0, 3))
        assert meta_tile.bbox == pytest.approx((-180.0, -90, 180.0, 90.0))
        assert meta_tile.size == (724, 362)
        assert meta_tile.tile_patterns == [
            ((0, 1, 3), (0, -149)),
            ((1, 1, 3), (256, -149)),
            ((2, 1, 3), (512, -149)),
            ((0, 0, 3), (0, 107)),
            ((1, 0, 3), (256, 107)),
            ((2, 0, 3), (512, 107)),
        ]

    def test_metatile_non_default_meta_size(self):
        mgrid = MetaGrid(grid=self.grid, meta_size=(4, 2), meta_buffer=0)
        meta_tile = mgrid.meta_tile((4, 3, 6))
        assert meta_tile.bbox == pytest.approx((0.0, 0.0, 180.0, 90.0))
        assert meta_tile.size == (1024, 512)
        assert meta_tile.grid_size == (4, 2)
        assert meta_tile.tile_patterns == [
            ((4, 3, 6), (0, 0)),
            ((5, 3, 6), (256, 0)),
            ((6, 3, 6), (512, 0)),
            ((7, 3, 6), (768, 0)),
            ((4, 2, 6), (0, 256)),
            ((5, 2, 6), (256, 256)),
            ((6, 2, 6), (512, 256)),
            ((7, 2, 6), (768, 256)),
        ]


class TestMinimalMetaTile(object):
    def setup_method(self):
        self.mgrid = MetaGrid(
            grid=tile_grid("EPSG:4326"), meta_size=(2, 2), meta_buffer=10
        )

    def test_minimal_tiles(self):
        sgrid = self.mgrid.minimal_meta_tile([(0, 0, 2), (1, 0, 2)])
        assert sgrid.grid_size == (2, 1)
        assert list(sgrid.tile_patterns) == [
            ((0, 0, 2), (0, 10)),
            ((1, 0, 2), (256, 10)),
        ]
        assert sgrid.bbox == (-180.0, -90.0, 3.515625, 3.515625)

    def test_minimal_tiles_fragmented(self):
        sgrid = self.mgrid.minimal_meta_tile(
            [
                (2, 3, 3),
                (1, 2, 3),
                (2, 1, 3),
            ]
        )

        assert sgrid.grid_size == (2, 3)
        assert list(sgrid.tile_patterns) == [
            ((1, 3, 3), (10, 0)),
            ((2, 3, 3), (266, 0)),
            ((1, 2, 3), (10, 256)),
            ((2, 2, 3), (266, 256)),
            ((1, 1, 3), (10, 512)),
            ((2, 1, 3), (266, 512)),
        ]
        assert sgrid.bbox == (-136.7578125, -46.7578125, -43.2421875, 90.0)

    def test_minimal_tiles_fragmented_ul(self):
        self.mgrid = MetaGrid(
            grid=tile_grid("EPSG:4326", origin="ul"), meta_size=(2, 2), meta_buffer=10
        )
        sgrid = self.mgrid.minimal_meta_tile(
            [
                (2, 0, 3),
                (1, 1, 3),
                (2, 2, 3),
            ]
        )

        assert sgrid.grid_size == (2, 3)
        assert list(sgrid.tile_patterns) == [
            ((1, 0, 3), (10, 0)),
            ((2, 0, 3), (266, 0)),
            ((1, 1, 3), (10, 256)),
            ((2, 1, 3), (266, 256)),
            ((1, 2, 3), (10, 512)),
            ((2, 2, 3), (266, 512)),
        ]
        assert sgrid.bbox == (-136.7578125, -46.7578125, -43.2421875, 90.0)


class TestMetaGridLevelMetaTiles(object):
    def setup_method(self):
        self.meta_grid = MetaGrid(TileGrid(), meta_size=(2, 2))

    def test_full_grid_0(self):
        bbox = (-20037508.34, -20037508.34, 20037508.34, 20037508.34)
        abbox, tile_grid, meta_tiles = self.meta_grid.get_affected_level_tiles(bbox, 0)
        meta_tiles = list(meta_tiles)
        assert_almost_equal_bbox(bbox, abbox)

        assert len(meta_tiles) == 1
        assert meta_tiles[0] == (0, 0, 0)

    def test_full_grid_2(self):
        bbox = (-20037508.34, -20037508.34, 20037508.34, 20037508.34)
        abbox, tile_grid, meta_tiles = self.meta_grid.get_affected_level_tiles(bbox, 2)
        meta_tiles = list(meta_tiles)
        assert_almost_equal_bbox(bbox, abbox)

        assert tile_grid == (2, 2)
        assert len(meta_tiles) == 4
        assert meta_tiles[0] == (0, 2, 2)
        assert meta_tiles[1] == (2, 2, 2)
        assert meta_tiles[2] == (0, 0, 2)
        assert meta_tiles[3] == (2, 0, 2)


class TestMetaGridLevelMetaTilesGeodetic(object):
    def setup_method(self):
        self.meta_grid = MetaGrid(TileGrid(is_geodetic=True), meta_size=(2, 2))

    def test_full_grid_2(self):
        bbox = (-180.0, -90.0, 180.0, 90)
        abbox, tile_grid, meta_tiles = self.meta_grid.get_affected_level_tiles(bbox, 2)
        meta_tiles = list(meta_tiles)
        assert_almost_equal_bbox(bbox, abbox)

        assert tile_grid == (2, 1)
        assert len(meta_tiles) == 2
        assert meta_tiles[0] == (0, 0, 2)
        assert meta_tiles[1] == (2, 0, 2)

    def test_partial_grid_3(self):
        bbox = (0.0, 5.0, 45, 40)
        abbox, tile_grid, meta_tiles = self.meta_grid.get_affected_level_tiles(bbox, 3)
        meta_tiles = list(meta_tiles)
        assert_almost_equal_bbox((0.0, 0.0, 90.0, 90.0), abbox)

        assert tile_grid == (1, 1)
        assert len(meta_tiles) == 1
        assert meta_tiles[0] == (4, 2, 3)


def assert_grid_size(grid, level, grid_size):
    print(grid.grid_sizes[level], "==", grid_size)
    assert grid.grid_sizes[level] == grid_size
    res = grid.resolutions[level]
    x, y = grid_size
    assert res * x * 256 >= grid.bbox[2] - grid.bbox[0]
    assert res * y * 256 >= grid.bbox[3] - grid.bbox[1]


class TileGridTest(object):
    def check_grid(self, level, grid_size):
        assert_grid_size(self.grid, level, grid_size)


class TestTileGridResolutions(object):
    def test_explicit_grid(self):
        grid = TileGrid(res=[0.1, 0.05, 0.01])
        assert grid.resolution(0) == 0.1
        assert grid.resolution(1) == 0.05
        assert grid.resolution(2) == 0.01

        assert grid.closest_level(0.00001) == 2

    def test_factor_grid(self):
        grid = TileGrid(is_geodetic=True, res=1 / 0.75, tile_size=(360, 180))
        assert grid.resolution(0) == 1.0
        assert grid.resolution(1) == 0.75
        assert grid.resolution(2) == 0.75 * 0.75

    def test_sqrt_grid(self):
        grid = TileGrid(is_geodetic=True, res="sqrt2", tile_size=(360, 180))
        assert grid.resolution(0) == 1.0
        assert grid.resolution(2) == pytest.approx(0.5)
        assert grid.resolution(4) == pytest.approx(0.25)


class TestWGS84TileGrid(object):
    def setup_method(self):
        self.grid = TileGrid(is_geodetic=True)

    def test_resolution(self):
        assert self.grid.resolution(0) == pytest.approx(1.40625)
        assert self.grid.resolution(1) == pytest.approx(1.40625 / 2)

    def test_bbox(self):
        assert self.grid.bbox == (-180.0, -90.0, 180.0, 90.0)

    def test_grid_size(self):
        assert self.grid.grid_sizes[0] == (1, 1)
        assert self.grid.grid_sizes[1] == (2, 1)
        assert self.grid.grid_sizes[2] == (4, 2)

    def test_affected_tiles(self):
        bbox, grid, tiles = self.grid.get_affected_tiles(
            (-180, -90, 180, 90), (512, 256)
        )
        assert bbox == (-180.0, -90.0, 180.0, 90.0)
        assert grid == (2, 1)
        assert list(tiles) == [(0, 0, 1), (1, 0, 1)]

    def test_affected_level_tiles(self):
        bbox, grid, tiles = self.grid.get_affected_level_tiles((-180, -90, 180, 90), 1)
        assert grid == (2, 1)
        assert bbox == (-180.0, -90.0, 180.0, 90.0)
        assert list(tiles) == [(0, 0, 1), (1, 0, 1)]
        bbox, grid, tiles = self.grid.get_affected_level_tiles((0, 0, 180, 90), 2)
        assert grid == (2, 1)
        assert bbox == (0.0, 0.0, 180.0, 90.0)
        assert list(tiles) == [(2, 1, 2), (3, 1, 2)]


class TestWGS83TileGridUL(object):
    def setup_method(self):
        self.grid = TileGrid(4326, bbox=(-180, -90, 180, 90), origin="ul")

    def test_resolution(self):
        assert self.grid.resolution(0) == pytest.approx(1.40625)
        assert self.grid.resolution(1) == pytest.approx(1.40625 / 2)

    def test_bbox(self):
        assert self.grid.bbox == (-180.0, -90.0, 180.0, 90.0)

    def test_tile_bbox(self):
        assert self.grid.tile_bbox((0, 0, 0)) == (-180.0, -270.0, 180.0, 90.0)
        assert self.grid.tile_bbox((0, 0, 0), limit=True) == (
            -180.0,
            -90.0,
            180.0,
            90.0,
        )
        assert self.grid.tile_bbox((0, 0, 1)) == (-180.0, -90.0, 0.0, 90.0)

    def test_tile(self):
        assert self.grid.tile(-170, -80, 0) == (0, 0, 0)
        assert self.grid.tile(-170, -80, 1) == (0, 0, 1)
        assert self.grid.tile(-170, -80, 2) == (0, 1, 2)

    def test_grid_size(self):
        assert self.grid.grid_sizes[0] == (1, 1)
        assert self.grid.grid_sizes[1] == (2, 1)
        assert self.grid.grid_sizes[2] == (4, 2)

    def test_affected_tiles(self):
        bbox, grid, tiles = self.grid.get_affected_tiles(
            (-180, -90, 180, 90), (512, 256)
        )
        assert bbox == (-180.0, -90.0, 180.0, 90.0)
        assert grid == (2, 1)
        assert list(tiles) == [(0, 0, 1), (1, 0, 1)]

        bbox, grid, tiles = self.grid.get_affected_tiles((-180, -90, 0, 90), (512, 512))
        assert bbox == (-180.0, -90.0, 0.0, 90.0)
        assert grid == (2, 2)
        assert list(tiles) == [(0, 0, 2), (1, 0, 2), (0, 1, 2), (1, 1, 2)]

    def test_affected_level_tiles(self):
        bbox, grid, tiles = self.grid.get_affected_level_tiles((-180, -90, 180, 90), 1)
        assert grid == (2, 1)
        assert bbox == (-180.0, -90.0, 180.0, 90.0)
        assert list(tiles) == [(0, 0, 1), (1, 0, 1)]
        bbox, grid, tiles = self.grid.get_affected_level_tiles((0, 0, 180, 90), 2)
        assert grid == (2, 1)
        assert list(tiles) == [(2, 0, 2), (3, 0, 2)]
        assert bbox == (0.0, 0.0, 180.0, 90.0)
        bbox, grid, tiles = self.grid.get_affected_level_tiles((0, -90, 180, 90), 2)
        assert grid == (2, 2)
        assert list(tiles) == [(2, 0, 2), (3, 0, 2), (2, 1, 2), (3, 1, 2)]
        assert bbox == (0.0, -90.0, 180.0, 90.0)


class TestGKTileGrid(TileGridTest):
    def setup_method(self):
        self.grid = TileGrid(SRS(31467), bbox=(3250000, 5230000, 3930000, 6110000))

    def test_bbox(self):
        assert self.grid.bbox == (3250000, 5230000, 3930000, 6110000)

    def test_resolution(self):
        res = self.grid.resolution(0)
        width = self.grid.bbox[2] - self.grid.bbox[0]
        height = self.grid.bbox[3] - self.grid.bbox[1]
        assert height == 880000.0 and width == 680000.0
        assert res == 880000.0 / 256

    def test_tile_bbox(self):
        tile_bbox = self.grid.tile_bbox((0, 0, 0))
        assert tile_bbox == (3250000.0, 5230000.0, 4130000.0, 6110000.0)

    def test_tile(self):
        x, y = 3450000, 5890000
        assert [self.grid.tile(x, y, level) for level in range(5)] == [
            (0, 0, 0),
            (0, 1, 1),
            (0, 3, 2),
            (1, 6, 3),
            (3, 12, 4),
        ]

    @pytest.mark.parametrize(
        "level,grid_size",
        [
            (0, (1, 1)),
            (1, (2, 2)),
            (2, (4, 4)),
            (3, (7, 8)),
        ],
    )
    def test_grids(self, level, grid_size):
        self.check_grid(level, grid_size)

    def test_closest_level(self):
        assert self.grid.closest_level(880000.0 / 256) == 0
        assert self.grid.closest_level(600000.0 / 256) == 1
        assert self.grid.closest_level(440000.0 / 256) == 1
        assert self.grid.closest_level(420000.0 / 256) == 1

    def test_adjacent_tile_bbox(self):
        t1 = self.grid.tile_bbox((0, 0, 1))
        t2 = self.grid.tile_bbox((1, 0, 1))
        t3 = self.grid.tile_bbox((0, 1, 1))
        assert t1[1] == t2[1]
        assert t1[3] == t2[3]
        assert t1[2] == t2[0]
        assert t1[0] == t3[0]
        assert t1[3] == t3[1]


class TestGKTileGridUL(TileGridTest):
    """
    Custom grid with ul origin.
    """

    def setup_method(self):
        self.grid = TileGrid(
            SRS(31467),
            bbox=(3300000, 5300000, 3900000, 6000000),
            origin="ul",
            res=[1500, 1000, 500, 300, 150, 100],
        )

    def test_bbox(self):
        assert self.grid.bbox == (3300000, 5300000, 3900000, 6000000)

    def test_tile_bbox(self):
        assert self.grid.tile_bbox((0, 0, 0)) == (
            3300000.0,
            5616000.0,
            3684000.0,
            6000000.0,
        )
        assert self.grid.tile_bbox((1, 0, 0)) == (
            3684000.0,
            5616000.0,
            4068000.0,
            6000000.0,
        )
        assert self.grid.tile_bbox((1, 1, 0)) == (
            3684000.0,
            5232000.0,
            4068000.0,
            5616000.0,
        )

    def test_tile(self):
        x, y = 3310000, 5990000
        assert self.grid.tile(x, y, 0) == (0, 0, 0)
        assert self.grid.tile(x, y, 1) == (0, 0, 1)
        assert self.grid.tile(x, y, 2) == (0, 0, 2)

        x, y = 3890000, 5310000
        assert self.grid.tile(x, y, 0) == (1, 1, 0)
        assert self.grid.tile(x, y, 1) == (2, 2, 1)
        assert self.grid.tile(x, y, 2) == (4, 5, 2)

    def test_grids(self):
        assert_grid_size(self.grid, 0, (2, 2))
        assert_grid_size(self.grid, 1, (3, 3))
        assert_grid_size(self.grid, 2, (5, 6))

    def test_closest_level(self):
        assert self.grid.closest_level(1500) == 0
        assert self.grid.closest_level(1000) == 1
        assert self.grid.closest_level(900) == 1
        assert self.grid.closest_level(600) == 2

    def test_adjacent_tile_bbox(self):
        t1 = self.grid.tile_bbox((0, 0, 1))
        t2 = self.grid.tile_bbox((1, 0, 1))
        t3 = self.grid.tile_bbox((0, 1, 1))
        assert t1[1] == t2[1]
        assert t1[3] == t2[3]
        assert t1[2] == t2[0]
        assert t1[0] == t3[0]
        assert t1[1] == t3[3]


class TestClosestLevelTinyResFactor(object):
    def setup_method(self):
        self.grid = TileGrid(
            SRS(31467),
            bbox=[420000, 30000, 900000, 350000],
            origin="ul",
            res=[
                4000,
                3750,
                3500,
                3250,
                3000,
                2750,
                2500,
                2250,
                2000,
                1750,
                1500,
                1250,
                1000,
                750,
                650,
                500,
                250,
                100,
                50,
                20,
                10,
                5,
                2.5,
                2,
                1.5,
                1,
                0.5,
            ],
        )

    def test_closest_level(self):
        assert self.grid.closest_level(5000) == 0
        assert self.grid.closest_level(4000) == 0
        assert self.grid.closest_level(3750) == 1
        assert self.grid.closest_level(3500) == 2
        assert self.grid.closest_level(3250) == 3
        assert self.grid.closest_level(3000) == 4


class TestOrigins(object):
    def test_basic(self):
        grid = tile_grid(4326, bbox=(-180, -90, 180, 90), origin="ll")
        assert grid.supports_access_with_origin("ll")
        assert not grid.supports_access_with_origin("ul")

        grid = tile_grid(4326, bbox=(-180, -90, 180, 90), origin="ul")
        assert not grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")

    def test_basic_no_level_zero(self):
        grid = tile_grid(
            4326, bbox=(-180, -90, 180, 90), origin="ll", min_res=360 / 256 / 2
        )
        assert grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")

        grid = tile_grid(
            4326, bbox=(-180, -90, 180, 90), origin="ul", min_res=360 / 256 / 2
        )
        assert grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")

    def test_basic_mixed_name(self):
        grid = tile_grid(4326, bbox=(-180, -90, 180, 90), origin="ll")
        assert grid.supports_access_with_origin("sw")
        assert not grid.supports_access_with_origin("nw")

        grid = tile_grid(4326, bbox=(-180, -90, 180, 90), origin="ul")
        assert not grid.supports_access_with_origin("sw")
        assert grid.supports_access_with_origin("nw")

    def test_custom_with_match(self):
        # height is divisible by res*tile_size
        grid = tile_grid(4326, bbox=(0, 0, 1024, 1024), origin="ll", min_res=1)
        assert grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")

        grid = tile_grid(4326, bbox=(0, 0, 1024, 1024), origin="ul", min_res=1)
        assert grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")

    def test_custom_without_match(self):
        # height is not divisible by res*tile_size
        grid = tile_grid(4326, bbox=(0, 0, 1024, 1000), origin="ll", min_res=1)
        assert grid.supports_access_with_origin("ll")
        assert not grid.supports_access_with_origin("ul")

        grid = tile_grid(4326, bbox=(0, 0, 1024, 1000), origin="ul", min_res=1)
        assert not grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")

    def test_custom_res_with_match(self):
        grid = tile_grid(4326, bbox=(0, 0, 1024, 1024), origin="ll", res=[1, 0.5, 0.25])
        assert grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")

        grid = tile_grid(4326, bbox=(0, 0, 1024, 1024), origin="ul", res=[1, 0.5, 0.25])
        assert grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")

    def test_custom_res_without_match(self):
        grid = tile_grid(4326, bbox=(0, 0, 1024, 1023), origin="ll", res=[1, 0.5, 0.25])
        assert grid.supports_access_with_origin("ll")
        assert not grid.supports_access_with_origin("ul")

        grid = tile_grid(4326, bbox=(0, 0, 1024, 1023), origin="ul", res=[1, 0.5, 0.25])
        assert not grid.supports_access_with_origin("ll")
        assert grid.supports_access_with_origin("ul")


class TestFixedResolutionsTileGrid(TileGridTest):
    def setup_method(self):
        self.res = [1000.0, 500.0, 200.0, 100.0, 50.0, 20.0, 5.0]
        bbox = (3250000, 5230000, 3930000, 6110000)
        self.grid = TileGrid(SRS(31467), bbox=bbox, res=self.res)

    def test_resolution(self):
        for level, res in enumerate(self.res):
            assert res == self.grid.resolution(level)

    def test_closest_level(self):
        assert self.grid.closest_level(2000) == 0
        assert self.grid.closest_level(1000) == 0
        assert self.grid.closest_level(950) == 0
        assert self.grid.closest_level(210) == 2

    def test_affected_tiles(self):
        req_bbox = (3250000, 5230000, 3930000, 6110000)
        self.grid.max_shrink_factor = 10
        bbox, grid_size, tiles = self.grid.get_affected_tiles(req_bbox, (256, 256))
        assert bbox == (
            req_bbox[0],
            req_bbox[1],
            req_bbox[0] + 1000 * 256 * 3,
            req_bbox[1] + 1000 * 256 * 4,
        )
        assert grid_size == (3, 4)
        tiles = list(tiles)
        assert tiles == [
            (0, 3, 0),
            (1, 3, 0),
            (2, 3, 0),
            (0, 2, 0),
            (1, 2, 0),
            (2, 2, 0),
            (0, 1, 0),
            (1, 1, 0),
            (2, 1, 0),
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
        ]

    def test_affected_tiles_2(self):
        req_bbox = (3250000, 5230000, 3930000, 6110000)
        self.grid.max_shrink_factor = 2.0
        try:
            bbox, grid_size, tiles = self.grid.get_affected_tiles(req_bbox, (256, 256))
        except NoTiles:
            pass
        else:
            assert False, "got no exception"

    @pytest.mark.parametrize(
        "level,grid_size",
        [
            (0, (3, 4)),
            (1, (6, 7)),
            (2, (14, 18)),
        ],
    )
    def test_grids(self, level, grid_size):
        self.check_grid(level, grid_size)

    def test_tile_bbox(self):
        tile_bbox = self.grid.tile_bbox((0, 0, 0))  # w: 1000x256
        assert tile_bbox == (3250000.0, 5230000.0, 3506000.0, 5486000.0)
        tile_bbox = self.grid.tile_bbox((0, 0, 1))  # w: 500x256
        assert tile_bbox == (3250000.0, 5230000.0, 3378000.0, 5358000.0)
        tile_bbox = self.grid.tile_bbox((0, 0, 2))  # w: 200x256
        assert tile_bbox == (3250000.0, 5230000.0, 3301200.0, 5281200.0)


class TestGeodeticTileGrid(TileGridTest):
    def setup_method(self):
        self.grid = TileGrid(
            is_geodetic=True,
        )

    def test_auto_resolution(self):
        grid = TileGrid(is_geodetic=True, bbox=(-10, 30, 10, 40), tile_size=(20, 20))
        tile_bbox = grid.tile_bbox((0, 0, 0))
        assert tile_bbox == (-10, 30, 10, 50)
        assert grid.resolution(0) == 1.0

    @pytest.mark.parametrize(
        "level,grid_size",
        [
            (0, (1, 1)),
            (1, (2, 1)),
            (2, (4, 2)),
        ],
    )
    def test_grids(self, level, grid_size):
        self.check_grid(level, grid_size)

    def test_adjacent_tile_bbox(self):
        grid = TileGrid(is_geodetic=True, bbox=(-10, 30, 10, 40), tile_size=(20, 20))
        t1 = grid.tile_bbox((0, 0, 2))
        t2 = grid.tile_bbox((1, 0, 2))
        t3 = grid.tile_bbox((0, 1, 2))
        assert t1[1] == t2[1]
        assert t1[3] == t2[3]
        assert t1[2] == t2[0]
        assert t1[0] == t3[0]
        assert t1[2] == t3[2]
        assert t1[3] == t3[1]

    def test_w_resolution(self):
        res = [1, 0.5, 0.2]
        grid = TileGrid(
            is_geodetic=True, bbox=(-10, 30, 10, 40), tile_size=(20, 20), res=res
        )
        assert grid.grid_sizes[0] == (1, 1)
        assert grid.grid_sizes[1] == (2, 1)
        assert grid.grid_sizes[2] == (5, 3)

    def test_tile(self):
        assert self.grid.tile(-180, -90, 0) == (0, 0, 0)
        assert self.grid.tile(-180, -90, 1) == (0, 0, 1)
        assert self.grid.tile(-180, -90, 2) == (0, 0, 2)
        assert self.grid.tile(180 - 0.001, 90 - 0.001, 0) == (0, 0, 0)
        assert self.grid.tile(10, 50, 1) == (1, 0, 1)

    def test_affected_tiles(self):
        bbox, grid_size, tiles = self.grid.get_affected_tiles(
            (-45, -45, 45, 45), (512, 512)
        )
        assert self.grid.grid_sizes[3] == (8, 4)
        assert bbox == (-45.0, -45.0, 45.0, 45.0)
        assert grid_size == (2, 2)
        tiles = list(tiles)
        assert tiles == [(3, 2, 3), (4, 2, 3), (3, 1, 3), (4, 1, 3)]


class TestTileGrid(object):
    def test_tile_out_of_grid_bounds(self):
        grid = TileGrid(is_geodetic=True)
        assert grid.tile(-180.01, 50, 1) == (-1, 0, 1)

    def test_affected_tiles_out_of_grid_bounds(self):
        grid = TileGrid()
        # bbox from open layers
        req_bbox = (
            -30056262.509599999,
            -10018754.170400001,
            -20037508.339999996,
            -0.00080000050365924835,
        )
        bbox, grid_size, tiles = grid.get_affected_tiles(req_bbox, (256, 256))
        assert_almost_equal_bbox(bbox, req_bbox, 1)
        assert grid_size == (1, 1)
        tiles = list(tiles)
        assert tiles == [None]

    def test_broken_bbox(self):
        grid = TileGrid()
        req_bbox = (
            -20000855.0573254,
            2847125.18913603,
            -19329367.42767611,
            4239924.78564583,
        )
        try:
            grid.get_affected_tiles(req_bbox, (256, 256), req_srs=SRS(31467))
        except TransformationError:
            pass
        else:
            assert False, "Expected TransformationError"


class TestTileGridThreshold(object):
    def test_lower_bound(self):
        # thresholds near the next lower res value
        grid = TileGrid(res=[1000, 500, 250, 100, 50], threshold_res=[300, 110])
        grid.stretch_factor = 1.1
        assert grid.closest_level(1100) == 0
        # regular transition (w/stretchfactor)
        assert grid.closest_level(950) == 0
        assert grid.closest_level(800) == 1
        assert grid.closest_level(500) == 1
        # transition at threshold
        assert grid.closest_level(301) == 1
        assert grid.closest_level(300) == 2
        assert grid.closest_level(250) == 2
        # transition at threshold
        assert grid.closest_level(111) == 2
        assert grid.closest_level(110) == 3
        assert grid.closest_level(100) == 3
        # regular transition (w/stretchfactor)
        assert grid.closest_level(92) == 3
        assert grid.closest_level(90) == 4

    def test_upper_bound(self):
        # thresholds near the next upper res value (within threshold)
        grid = TileGrid(res=[1000, 500, 250, 100, 50], threshold_res=[495, 240])
        grid.stretch_factor = 1.1
        assert grid.closest_level(1100) == 0
        # regular transition (w/stretchfactor)
        assert grid.closest_level(950) == 0
        assert grid.closest_level(800) == 1
        assert grid.closest_level(500) == 1
        # transition at threshold
        assert grid.closest_level(496) == 1
        assert grid.closest_level(495) == 2
        assert grid.closest_level(250) == 2
        # transition at threshold (within strechfactor)
        assert grid.closest_level(241) == 2
        assert grid.closest_level(240) == 3
        assert grid.closest_level(100) == 3
        # regular transition (w/stretchfactor)
        assert grid.closest_level(92) == 3
        assert grid.closest_level(90) == 4

    def test_above_first_res(self):
        grid = TileGrid(res=[1000, 500, 250, 100, 50], threshold_res=[1100, 750])
        grid.stretch_factor = 1.1
        assert grid.closest_level(1200) == 0
        assert grid.closest_level(1100) == 0
        assert grid.closest_level(1000) == 0
        assert grid.closest_level(800) == 0
        assert grid.closest_level(750.1) == 0
        assert grid.closest_level(750) == 1


class TestCreateTileList(object):
    def test(self):
        xs = list(range(-1, 2))
        ys = list(range(-2, 3))
        grid_size = (1, 2)
        tiles = list(_create_tile_list(xs, ys, 3, grid_size))

        expected = [
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            (0, 0, 3),
            None,
            None,
            (0, 1, 3),
            None,
            None,
            None,
            None,
        ]
        assert expected == tiles

    def _create_tile_list(self, xs, ys, level, grid_size):
        x_limit = grid_size[0]
        y_limit = grid_size[1]
        for y in ys:
            for x in xs:
                if x < 0 or y < 0 or x >= x_limit or y >= y_limit:
                    yield None
                else:
                    yield x, y, level


class TestBBOXIntersects(object):
    def test_no_intersect(self):
        b1 = (0, 0, 10, 10)
        b2 = (20, 20, 30, 30)
        assert not bbox_intersects(b1, b2)
        assert not bbox_intersects(b2, b1)

    def test_no_intersect_only_vertical(self):
        b1 = (0, 0, 10, 10)
        b2 = (20, 0, 30, 10)
        assert not bbox_intersects(b1, b2)
        assert not bbox_intersects(b2, b1)

    def test_no_intersect_touch_point(self):
        b1 = (0, 0, 10, 10)
        b2 = (10, 10, 20, 20)
        assert not bbox_intersects(b1, b2)
        assert not bbox_intersects(b2, b1)

    def test_no_intersect_touch_side(self):
        b1 = (0, 0, 10, 10)
        b2 = (0, 10, 10, 20)
        assert not bbox_intersects(b1, b2)
        assert not bbox_intersects(b2, b1)

    def test_full_contains(self):
        b1 = (0, 0, 10, 10)
        b2 = (2, 2, 8, 8)
        assert bbox_intersects(b1, b2)
        assert bbox_intersects(b2, b1)

    def test_overlap(self):
        b1 = (0, 0, 10, 10)
        b2 = (-5, -5, 5, 5)
        assert bbox_intersects(b1, b2)
        assert bbox_intersects(b2, b1)


class TestBBOXContains(object):
    def test_no_intersect(self):
        b1 = (0, 0, 10, 10)
        b2 = (20, 20, 30, 30)
        assert not bbox_contains(b1, b2)
        assert not bbox_contains(b2, b1)

    def test_no_intersect_only_vertical(self):
        b1 = (0, 0, 10, 10)
        b2 = (20, 0, 30, 10)
        assert not bbox_contains(b1, b2)
        assert not bbox_contains(b2, b1)

    def test_no_intersect_touch_point(self):
        b1 = (0, 0, 10, 10)
        b2 = (10, 10, 20, 20)
        assert not bbox_contains(b1, b2)
        assert not bbox_contains(b2, b1)

    def test_no_intersect_touch_side(self):
        b1 = (0, 0, 10, 10)
        b2 = (0, 10, 10, 20)
        assert not bbox_contains(b1, b2)
        assert not bbox_contains(b2, b1)

    def test_full_contains(self):
        b1 = (0, 0, 10, 10)
        b2 = (2, 2, 8, 8)
        assert bbox_contains(b1, b2)
        assert not bbox_contains(b2, b1)

    def test_contains_touch(self):
        b1 = (0, 0, 10, 10)
        b2 = (0, 0, 8, 8)
        assert bbox_contains(b1, b2)
        assert not bbox_contains(b2, b1)

    def test_overlap(self):
        b1 = (0, 0, 10, 10)
        b2 = (-5, -5, 5, 5)
        assert not bbox_contains(b1, b2)
        assert not bbox_contains(b2, b1)


def assert_almost_equal_bbox(bbox1, bbox2, rel=0.01):
    assert bbox1 == pytest.approx(bbox2, rel=rel)


class TestResolutionRange(object):
    def test_meter(self):
        res_range = ResolutionRange(1000, 10)
        assert not res_range.contains([0, 0, 100000, 100000], (10, 10), SRS(900913))
        assert not res_range.contains([0, 0, 100000, 100000], (99, 99), SRS(900913))
        # min is exclusive but there is a delta
        assert res_range.contains([0, 0, 100000, 100000], (100, 100), SRS(900913))
        assert res_range.contains([0, 0, 100000, 100000], (1000, 1000), SRS(900913))
        # max is inclusive
        assert res_range.contains([0, 0, 100000, 100000], (10000, 10000), SRS(900913))
        assert not res_range.contains(
            [0, 0, 100000, 100000], (10001, 10001), SRS(900913)
        )

    def test_deg(self):
        res_range = ResolutionRange(100000, 1000)
        assert not res_range.contains([0, 0, 10, 10], (10, 10), SRS(4326))
        assert not res_range.contains([0, 0, 10, 10], (11, 11), SRS(4326))
        assert res_range.contains([0, 0, 10, 10], (12, 12), SRS(4326))
        assert res_range.contains([0, 0, 10, 10], (100, 100), SRS(4326))
        assert res_range.contains([0, 0, 10, 10], (1000, 1000), SRS(4326))
        assert res_range.contains([0, 0, 10, 10], (1100, 1100), SRS(4326))
        assert not res_range.contains([0, 0, 10, 10], (1200, 1200), SRS(4326))

    def test_no_min(self):
        res_range = ResolutionRange(None, 10)
        assert res_range.contains([0, 0, 100000, 100000], (1, 1), SRS(900913))
        assert res_range.contains([0, 0, 100000, 100000], (10, 10), SRS(900913))
        assert res_range.contains([0, 0, 100000, 100000], (99, 99), SRS(900913))
        assert res_range.contains([0, 0, 100000, 100000], (100, 100), SRS(900913))
        assert res_range.contains([0, 0, 100000, 100000], (1000, 1000), SRS(900913))
        # max is inclusive
        assert res_range.contains([0, 0, 100000, 100000], (10000, 10000), SRS(900913))
        assert not res_range.contains(
            [0, 0, 100000, 100000], (10001, 10001), SRS(900913)
        )

    def test_no_max(self):
        res_range = ResolutionRange(1000, None)
        assert not res_range.contains([0, 0, 100000, 100000], (10, 10), SRS(900913))
        assert not res_range.contains([0, 0, 100000, 100000], (99, 99), SRS(900913))
        # min is exclusive but there is a delta
        assert res_range.contains([0, 0, 100000, 100000], (100, 100), SRS(900913))
        assert res_range.contains([0, 0, 100000, 100000], (1000, 1000), SRS(900913))
        assert res_range.contains([0, 0, 100000, 100000], (10000, 10000), SRS(900913))
        assert res_range.contains([0, 0, 100000, 100000], (10001, 10001), SRS(900913))
        assert res_range.contains(
            [0, 0, 100000, 100000], (1000000, 100000), SRS(900913)
        )

    def test_none(self):
        res_range = resolution_range(None, None)
        assert res_range is None

    def test_from_scale(self):
        res_range = resolution_range(max_scale=1e6, min_scale=1e3)
        assert res_range.min_res == pytest.approx(280)
        assert res_range.max_res == pytest.approx(0.28)

    @pytest.mark.parametrize(
        "min_res,max_res,max_scale,min_scale",
        [
            [10, None, 10, None],
            [10, 20, 10, None],
            [10, None, 10, 20],
            [10, 20, 10, 20],
        ],
    )
    def test_invalid_combination(self, min_res, max_res, max_scale, min_scale):
        with pytest.raises(ValueError):
            resolution_range(min_res, max_res, max_scale, min_scale)

    def test_wrong_order_res(self):
        with pytest.raises(AssertionError):
            resolution_range(min_res=10, max_res=100)

    def test_wrong_order_scale(self):
        with pytest.raises(AssertionError):
            resolution_range(min_scale=100, max_scale=10)

    def test_merge_resolutions(self):
        res_range = merge_resolution_range(
            ResolutionRange(None, 10), ResolutionRange(1000, None)
        )
        assert res_range is None

        res_range = merge_resolution_range(
            ResolutionRange(10000, 10), ResolutionRange(1000, None)
        )
        assert res_range.min_res == 10000
        assert res_range.max_res is None

        res_range = merge_resolution_range(
            ResolutionRange(10000, 10), ResolutionRange(1000, 1)
        )
        assert res_range.min_res == 10000
        assert res_range.max_res == 1

        res_range = merge_resolution_range(
            ResolutionRange(10000, 10), ResolutionRange(None, None)
        )
        assert res_range is None

        res_range = merge_resolution_range(None, ResolutionRange(None, None))
        assert res_range is None

        res_range = merge_resolution_range(ResolutionRange(10000, 10), None)
        assert res_range is None

    def test_eq(self):
        assert resolution_range(None, None) == resolution_range(None, None)
        assert resolution_range(None, 100) == resolution_range(None, 100.0)
        assert resolution_range(None, 100) != resolution_range(None, 100.1)
        assert resolution_range(1000, 100) == resolution_range(1000, 100)
        assert resolution_range(1000, 100) == resolution_range(1000.0, 100)
        assert resolution_range(1000, 100) != resolution_range(1000.1, 100)


class TestGridSubset(object):
    def test_different_srs(self):
        g1 = tile_grid(SRS(4326))
        g2 = tile_grid(SRS(3857))

        assert not g1.is_subset_of(g2)

    def test_same_grid(self):
        g1 = tile_grid(SRS(900913))

        assert g1.is_subset_of(g1)

    def test_similar_srs(self):
        g1 = tile_grid(SRS(900913))
        g2 = tile_grid(SRS(3857))

        assert g1.is_subset_of(g2)

    def test_less_levels(self):
        g1 = tile_grid(SRS(3857), num_levels=10)
        g2 = tile_grid(SRS(3857))

        assert g1.is_subset_of(g2)

    def test_more_levels(self):
        g1 = tile_grid(SRS(3857))
        g2 = tile_grid(SRS(3857), num_levels=10)

        assert not g1.is_subset_of(g2)

    def test_res_subset(self):
        g1 = tile_grid(SRS(3857), res=[50000, 10000, 100, 1])
        g2 = tile_grid(SRS(3857), res=[100000, 50000, 10000, 1000, 100, 10, 1, 0.5])

        assert g1.tile_bbox((0, 0, 0)) != g2.tile_bbox((0, 0, 0))
        assert g1.is_subset_of(g2)

        g1 = tile_grid(
            SRS(3857),
            bbox=[0, 0, 20037508.342789244, 20037508.342789244],
            min_res=78271.51696402048,
            num_levels=18,
        )
        g2 = tile_grid(SRS(3857), origin="nw")
        assert g1.is_subset_of(g2)

    def test_subbbox(self):
        g2 = tile_grid(SRS(4326))
        g1 = tile_grid(
            SRS(4326), num_levels=10, min_res=g2.resolutions[3], bbox=(0, 0, 180, 90)
        )

        assert g1.is_subset_of(g2)

    def test_incompatible_subbbox(self):
        g2 = tile_grid(SRS(4326))
        g1 = tile_grid(
            SRS(4326), min_res=g2.resolutions[3], num_levels=10, bbox=(-10, 0, 180, 90)
        )

        assert not g1.is_subset_of(g2)

    def test_tile_size(self):
        g1 = tile_grid(SRS(4326), tile_size=(128, 128))
        g2 = tile_grid(SRS(4326))

        assert not g1.is_subset_of(g2)

    def test_non_matching_bboxfor_origins(self):
        g1 = tile_grid(
            SRS(21781), bbox=[420000, 30000, 900000, 360000], res=[250], origin="nw"
        )
        g2 = tile_grid(
            SRS(21781), bbox=[420000, 30000, 900000, 360000], res=[250], origin="sw"
        )

        assert not g1.is_subset_of(g2)

    def test_no_tile_errors(self):
        # g1 is not a subset, check that we don't get any NoTile errors
        g1 = tile_grid(SRS(3857), res=[100000, 50000, 10000, 1000, 100, 10, 1, 0.5])
        g2 = tile_grid(SRS(3857), res=[100, 1])

        assert not g1.is_subset_of(g2)


WebMercatorQuad = {
    "id": "WebMercatorQuad",
    "title": "WebMercatorQuad",
    "uri": "http://www.opengis.net/def/tilematrixset/OGC/1.0/WebMercatorQuad",
    "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
    "orderedAxes": ["E", "N"],
    "wellKnownScaleSet": "http://www.opengis.net/def/wkss/OGC/1.0/GoogleMapsCompatible",
    "tileMatrices": [
        {
            "id": "0",
            "scaleDenominator": 559082264.028718,
            "cellSize": 156543.03392804,
            "cornerOfOrigin": "topLeft",
            "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
            "matrixWidth": 1,
            "matrixHeight": 1,
            "tileWidth": 256,
            "tileHeight": 256,
        },
        {
            "id": "1",
            "scaleDenominator": 279541132.01436,
            "cellSize": 78271.5169640205,
            "cornerOfOrigin": "topLeft",
            "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
            "matrixWidth": 2,
            "matrixHeight": 2,
            "tileWidth": 256,
            "tileHeight": 256,
        },
        {
            "id": "2",
            "scaleDenominator": 139770566.00718,
            "cellSize": 39135.7584820102,
            "cornerOfOrigin": "topLeft",
            "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
            "matrixWidth": 4,
            "matrixHeight": 4,
            "tileWidth": 256,
            "tileHeight": 256,
        },
    ],
}


def test_tile_grid_from_ogc_tile_matrix_set_WebMercatorQuad():
    tile_grid = tile_grid_from_ogc_tile_matrix_set(WebMercatorQuad)
    assert tile_grid.srs.srs_code == "EPSG:3857"
    assert tile_grid.tile_size == (256, 256)
    assert tile_grid.origin == "ul"
    assert tile_grid.name == "WebMercatorQuad"
    assert tile_grid.is_geodetic is False
    assert tile_grid.bbox == [
        -20037508.3427892,
        -20037508.342789043,
        20037508.342789043,
        20037508.3427892,
    ]
    assert tile_grid.levels == 3
    assert len(tile_grid.grid_sizes) == 3
    assert tile_grid.grid_sizes[0] == (1, 1)
    assert tile_grid.grid_sizes[1] == (2, 2)
    assert tile_grid.grid_sizes[2] == (4, 4)


LatLonGrid = {
    "id": "LatLonGrid",
    "title": "LatLonGrid",
    "crs": "http://www.opengis.net/def/crs/EPSG/0/4326",
    "orderedAxes": ["Lat", "Lon"],
    "tileMatrices": [
        {
            "id": "0",
            "scaleDenominator": 388251.572242165,
            "cellSize": 0.0009765625,
            "cornerOfOrigin": "topLeft",
            "pointOfOrigin": [90, -180],
            "matrixWidth": 360,
            "matrixHeight": 180,
            "tileWidth": 1024,
            "tileHeight": 1024,
        }
    ],
}


def test_tile_grid_from_ogc_tile_matrix_set_LatLonGrid():
    tile_grid = tile_grid_from_ogc_tile_matrix_set(LatLonGrid)
    assert tile_grid.srs.srs_code == "EPSG:4326"
    assert tile_grid.tile_size == (1024, 1024)
    assert tile_grid.origin == "ul"
    assert tile_grid.name == "LatLonGrid"
    assert tile_grid.is_geodetic is True
    assert tile_grid.bbox == [-180, -90, 180, 90]
    assert tile_grid.levels == 1
    assert len(tile_grid.grid_sizes) == 1
    assert tile_grid.grid_sizes[0] == (360, 180)


def test_tile_grid_from_ogc_tile_matrix_set_errors():
    d = copy.deepcopy(WebMercatorQuad)
    d["tileMatrices"][0]["variableMatrixWidths"] = []
    with pytest.raises(
        UnsupportedException,
        match="Tile matrix with variableMatrixWidths are not supported",
    ):
        tile_grid_from_ogc_tile_matrix_set(d)

    d = copy.deepcopy(WebMercatorQuad)
    d["tileMatrices"][0]["cornerOfOrigin"] = "invalid"
    with pytest.raises(UnsupportedException, match="Invalid cornerOfOrigin=invalid"):
        tile_grid_from_ogc_tile_matrix_set(d)

    d = copy.deepcopy(WebMercatorQuad)
    d["tileMatrices"][1]["tileWidth"] = 512
    with pytest.raises(
        UnsupportedException,
        match="Tile matrix set with varying tileWidth depending on tile matrix",
    ):
        tile_grid_from_ogc_tile_matrix_set(d)

    d = copy.deepcopy(WebMercatorQuad)
    d["tileMatrices"][1]["tileHeight"] = 512
    with pytest.raises(
        UnsupportedException,
        match="Tile matrix set with varying tileHeight depending on tile matrix",
    ):
        tile_grid_from_ogc_tile_matrix_set(d)

    d = copy.deepcopy(WebMercatorQuad)
    d["tileMatrices"][1]["pointOfOrigin"][0] = 0
    with pytest.raises(
        UnsupportedException,
        match="Tile matrix set with varying pointOfOrigin depending on tile matrix",
    ):
        tile_grid_from_ogc_tile_matrix_set(d)

    d = copy.deepcopy(WebMercatorQuad)
    d["tileMatrices"][1]["pointOfOrigin"][1] = 0
    with pytest.raises(
        UnsupportedException,
        match="Tile matrix set with varying pointOfOrigin depending on tile matrix",
    ):
        tile_grid_from_ogc_tile_matrix_set(d)



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division

import os
import re
import shlex
import sys
import optparse

import yaml

from mapproxy.srs import SRS
from mapproxy.config.coverage import load_coverage
from mapproxy.config.loader import (
    load_configuration, ConfigurationError,
    CacheConfiguration, GridConfiguration,
    ProxyConfiguration
)
from mapproxy.util.coverage import BBOXCoverage
from mapproxy.seed.util import ProgressLog, format_bbox
from mapproxy.seed.seeder import SeedTask, seed_task
from mapproxy.config import spec as conf_spec
from mapproxy.util.ext.dictspec.validator import validate, ValidationError

__all__ = [
    'SRS', 'load_coverage', 'load_configuration', 'ConfigurationError', 'CacheConfiguration', 'GridConfiguration',
    'ProxyConfiguration', 'BBOXCoverage']


def parse_levels(level_str):
    """
    >>> parse_levels('1,2,3,6')
    [1, 2, 3, 6]
    >>> parse_levels('1..6')
    [1, 2, 3, 4, 5, 6]
    >>> parse_levels('1..6, 8, 9, 13..14')
    [1, 2, 3, 4, 5, 6, 8, 9, 13, 14]
    """
    levels = set()
    for part in level_str.split(','):
        part = part.strip()
        if re.match(r'\d+..\d+', part):
            from_level, to_level = part.split('..')
            levels.update(list(range(int(from_level), int(to_level) + 1)))
        else:
            levels.add(int(part))

    return sorted(levels)


def resolve_source(source_name, conf):
    """
    Resolves the source with the given name.

    >>> config = ProxyConfiguration({'sources': {'mysource': {'type': 'wms'}}})
    >>> resolve_source('mysource', config)
    (<mapproxy.config.loader.WMSSourceConfiguration object at 0x...>, False)

    >>> config = ProxyConfiguration({'caches': {'mysource': {}}})
    >>> resolve_source('mysource', config)
    (<mapproxy.config.loader.CacheConfiguration object at 0x...>, True)

    >>> config = ProxyConfiguration({'caches': {'mysource': {'type': 'foo'}}})
    >>> resolve_source('nonexistingsource', config)
    (None, None)

    :param source_name: The name of the source to resolve.
    :param conf: The mapproxy config to resolve from.
    :rtype: (dict, bool) A tuple containing the resolved source and a boolean
        indicating if the resolved source is a cache (True), or a source (False).
    """
    resolved_source = conf.sources.get(source_name)
    if resolved_source is not None:
        return resolved_source, False
    resolved_source = conf.caches.get(source_name)
    if resolved_source is not None:
        return resolved_source, True
    return None, None


def parse_grid_definition(definition):
    """
    >>> sorted(parse_grid_definition("res=[10000,1000,100,10] srs=EPSG:4326 bbox=5,50,10,60").items())
    [('bbox', '5,50,10,60'), ('res', [10000, 1000, 100, 10]), ('srs', 'EPSG:4326')]
    """
    args = shlex.split(definition)
    grid_conf = {}
    for arg in args:
        key, value = arg.split('=')
        value = yaml.safe_load(value)
        grid_conf[key] = value

    validate(conf_spec.grid_opts, grid_conf)
    return grid_conf


def supports_tiled_access(mgr):
    if len(mgr.sources) == 1 and getattr(mgr.sources[0], 'supports_meta_tiles') is False:
        return True
    return False


def format_export_task(task, custom_grid):
    info = []
    if custom_grid:
        grid = "custom grid"
    else:
        grid = "grid '%s'" % task.md['grid_name']

    info.append("Exporting cache '%s' to '%s' with %s in %s" % (
        task.md['cache_name'], task.md['dest'], grid, task.grid.srs.srs_code))
    if task.coverage:
        info.append('  Limited to: %s (EPSG:4326)' % (format_bbox(task.coverage.extent.llbbox), ))
    info.append('  Levels: %s' % (task.levels, ))

    return '\n'.join(info)


def export_command(args=None):
    parser = optparse.OptionParser("%prog export [options] mapproxy_conf")
    parser.add_option("-f", "--mapproxy-conf", dest="mapproxy_conf",
                      help="MapProxy configuration")

    parser.add_option("-q", "--quiet",
                      action="count", dest="quiet", default=0,
                      help="reduce number of messages to stdout, repeat to disable progress output")

    parser.add_option("--source", dest="source",
                      help="source to export (source or cache)")

    parser.add_option("--grid",
                      help="grid for export. either the name of an existing grid or "
                      "the grid definition as a string")

    parser.add_option("--dest",
                      help="destination of the export (directory or filename)")

    parser.add_option("--type",
                      help="type of the export format")

    parser.add_option("--levels",
                      help="levels to export: e.g 1,2,3 or 1..10")

    parser.add_option("--fetch-missing-tiles", dest="fetch_missing_tiles",
                      action='store_true', default=False,
                      help="if missing tiles should be fetched from the sources")

    parser.add_option("--force",
                      action='store_true', default=False,
                      help="overwrite/append to existing --dest files/directories")

    parser.add_option("-n", "--dry-run",
                      action="store_true", default=False,
                      help="do not export, just print output")

    parser.add_option("-c", "--concurrency", type="int",
                      dest="concurrency", default=1,
                      help="number of parallel export processes")

    parser.add_option("--coverage",
                      help="the coverage for the export as a BBOX string, WKT file "
                      "or OGR datasource")
    parser.add_option("--srs",
                      help="the SRS of the coverage")
    parser.add_option("--where",
                      help="filter for OGR coverages")

    from mapproxy.script.util import setup_logging
    import logging
    setup_logging(logging.WARN)

    if args:
        args = args[1:]  # remove script name

    (options, args) = parser.parse_args(args)

    if not options.mapproxy_conf:
        if len(args) != 1:
            parser.print_help()
            sys.exit(1)
        else:
            options.mapproxy_conf = args[0]

    required_options = ['mapproxy_conf', 'grid', 'source', 'dest', 'levels']
    for required in required_options:
        if not getattr(options, required):
            print('ERROR: missing required option --%s' % required.replace('_', '-'), file=sys.stderr)
            parser.print_help()
            sys.exit(1)

    try:
        conf = load_configuration(options.mapproxy_conf)
    except IOError as e:
        print('ERROR: ', "%s: '%s'" % (e.strerror, e.filename), file=sys.stderr)
        sys.exit(2)
    except ConfigurationError as e:
        print(e, file=sys.stderr)
        print('ERROR: invalid configuration (see above)', file=sys.stderr)
        sys.exit(2)

    if '=' in options.grid:
        try:
            grid_conf = parse_grid_definition(options.grid)
        except ValidationError as ex:
            print('ERROR: invalid grid configuration', file=sys.stderr)
            for error in ex.errors:
                print(' ', error, file=sys.stderr)
            sys.exit(2)
        except ValueError:
            print('ERROR: invalid grid configuration', file=sys.stderr)
            sys.exit(2)
        options.grid = 'tmp_mapproxy_export_grid'
        grid_conf['name'] = options.grid
        custom_grid = True
        conf.grids[options.grid] = GridConfiguration(grid_conf, conf)
    else:
        custom_grid = False

    if os.path.exists(options.dest) and not options.force:
        print('ERROR: destination exists, remove first or use --force', file=sys.stderr)
        sys.exit(2)

    cache_conf = {
        'name': 'export',
        'grids': [options.grid],
        'sources': [options.source],
    }

    resolved_source, source_is_cache = resolve_source(options.source, conf)
    if source_is_cache:
        image_format = resolved_source.conf.get('format') or resolved_source.defaults.get('format')
        if image_format is not None:
            cache_conf['format'] = image_format
        if image_format == 'mixed':
            request_format = resolved_source.conf.get('request_format') or 'image/png'
            cache_conf['request_format'] = request_format

    if options.type == 'mbtile':
        cache_conf['cache'] = {
            'type': 'mbtiles',
            'filename': options.dest,
        }
    elif options.type == 'sqlite':
        cache_conf['cache'] = {
            'type': 'sqlite',
            'directory': options.dest,
        }
    elif options.type == 'geopackage':
        cache_conf['cache'] = {
            'type': 'geopackage',
            'filename': options.dest,
        }
    elif options.type == 'compact-v1':
        cache_conf['cache'] = {
            'type': 'compact',
            'version': 1,
            'directory': options.dest,
        }
    elif options.type == 'compact-v2':
        cache_conf['cache'] = {
            'type': 'compact',
            'version': 2,
            'directory': options.dest,
        }
    elif options.type in ('tc', 'mapproxy'):
        cache_conf['cache'] = {
            'type': 'file',
            'directory': options.dest,
        }
    elif options.type == 'arcgis':
        cache_conf['cache'] = {
            'type': 'file',
            'directory_layout': 'arcgis',
            'directory': options.dest,
        }
    elif options.type in ('tms', None):  # default
        cache_conf['cache'] = {
            'type': 'file',
            'directory_layout': 'tms',
            'directory': options.dest,
        }
    else:
        print('ERROR: unsupported --type %s' % (options.type, ), file=sys.stderr)
        sys.exit(2)

    if not options.fetch_missing_tiles:
        for source in conf.sources.values():
            source.conf['seed_only'] = True

    tile_grid, extent, mgr = CacheConfiguration(cache_conf, conf).caches()[0]

    levels = parse_levels(options.levels)
    if levels[-1] >= tile_grid.levels:
        print('ERROR: destination grid only has %d levels' % tile_grid.levels, file=sys.stderr)
        sys.exit(2)

    if options.srs:
        srs = SRS(options.srs)
    else:
        srs = tile_grid.srs

    if options.coverage:
        seed_coverage = load_coverage(
            {'datasource': options.coverage, 'srs': srs, 'where': options.where},
            base_path=os.getcwd())
    else:
        seed_coverage = BBOXCoverage(tile_grid.bbox, tile_grid.srs)

    if not supports_tiled_access(mgr):
        print('WARN: grids are incompatible. needs to scale/reproject tiles for export.', file=sys.stderr)

    md = dict(name='export', cache_name='cache', grid_name=options.grid, dest=options.dest)
    task = SeedTask(md, mgr, levels, refresh_timestamp=None, refresh_all=True, coverage=seed_coverage)

    print(format_export_task(task, custom_grid=custom_grid))

    logger = ProgressLog(verbose=options.quiet == 0, silent=options.quiet >= 2)
    try:
        seed_task(task, progress_logger=logger, dry_run=options.dry_run,
                  concurrency=options.concurrency)
    except KeyboardInterrupt:
        print('stopping...', file=sys.stderr)
        sys.exit(2)



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image, ImageDraw
from mapproxy.srs import SRS, make_lin_transf
from mapproxy.image import ImageSource
from mapproxy.image.opts import create_image
from mapproxy.util.geom import flatten_to_polygons


def mask_image_source_from_coverage(img_source, bbox, bbox_srs, coverage,
                                    image_opts=None):
    if image_opts is None:
        image_opts = img_source.image_opts
    img = img_source.as_image()
    img = mask_image(img, bbox, bbox_srs, coverage)
    result = create_image(img.size, image_opts)
    result.paste(img, (0, 0), img)
    return ImageSource(result, image_opts=image_opts)


def mask_image(img, bbox, bbox_srs, coverage):
    geom = mask_polygons(bbox, SRS(bbox_srs), coverage)
    mask = image_mask_from_geom(img.size, bbox, geom)
    img = img.convert('RGBA')
    img.paste((255, 255, 255, 0), (0, 0), mask)
    return img


def mask_polygons(bbox, bbox_srs, coverage):
    coverage = coverage.transform_to(bbox_srs)
    coverage = coverage.intersection(bbox, bbox_srs)
    return flatten_to_polygons(coverage.geom)


def image_mask_from_geom(size, bbox, polygons):
    mask = Image.new('L', size, 255)
    if len(polygons) == 0:
        return mask

    transf = make_lin_transf(bbox, (0, 0) + size)

    # use negative ~.1 pixel buffer
    buffer = -0.1 * min((bbox[2] - bbox[0]) / size[0], (bbox[3] - bbox[1]) / size[1])

    draw = ImageDraw.Draw(mask)

    def draw_polygon(p):
        draw.polygon([transf(coord) for coord in p.exterior.coords], fill=0)
        for ring in p.interiors:
            draw.polygon([transf(coord) for coord in ring.coords], fill=255)

    for p in polygons:
        # little bit smaller polygon does not include touched pixels outside coverage
        buffered = p.buffer(buffer, resolution=1, join_style=2)

        if buffered.is_empty:  # can be empty after negative buffer
            continue

        if buffered.geom_type == 'MultiPolygon':
            # negative buffer can turn polygon into multipolygon
            for p in buffered.geoms:
                draw_polygon(p)
        else:
            draw_polygon(buffered)

    return mask



# -:- encoding: UTF8 -:-
# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import pickle

import pytest

from mapproxy.srs import SRS
from mapproxy.request.base import url_decode, Request, NoCaseMultiDict, RequestParams
from mapproxy.request.tile import TMSRequest, tile_request, TileRequest
from mapproxy.request.wms import (
    wms_request,
    WMSMapRequest,
    WMSMapRequestParams,
    WMS111MapRequest,
    WMS100MapRequest,
    WMS130MapRequest,
    WMS111FeatureInfoRequest,
)
from mapproxy.request.arcgis import ArcGISRequest, ArcGISIdentifyRequest
from mapproxy.exception import RequestError
from mapproxy.request.wms.exception import (
    WMS111ExceptionHandler,
    WMSImageExceptionHandler,
    WMSBlankExceptionHandler,
)
from mapproxy.test.http import make_wsgi_env, assert_url_eq, assert_query_eq


class TestNoCaseMultiDict(object):

    def test_from_iterable(self):
        data = (("layers", "foo,bar"), ("laYERs", "baz"), ("crs", "EPSG:4326"))
        nc_dict = NoCaseMultiDict(data)
        print(nc_dict)

        for name in ("layers", "LAYERS", "lAYeRS"):
            assert name in nc_dict, name + " not found"
        assert nc_dict.get_all("layers") == ["foo,bar", "baz"]
        assert nc_dict.get_all("crs") == ["EPSG:4326"]

    def test_from_dict(self):
        data = [("layers", "foo,bar"), ("laYERs", "baz"), ("crs", "EPSG:4326")]
        nc_dict = NoCaseMultiDict(data)
        print(nc_dict)

        for name in ("layers", "LAYERS", "lAYeRS"):
            assert name in nc_dict, name + " not found"
        assert nc_dict.get_all("layers") == ["foo,bar", "baz"]
        assert nc_dict.get_all("crs") == ["EPSG:4326"]

    def test_iteritems(self):
        data = [("LAYERS", "foo,bar"), ("laYERs", "baz"), ("crs", "EPSG:4326")]
        nc_dict = NoCaseMultiDict(data)

        for key, values in nc_dict.iteritems():
            if key in ("LAYERS", "laYERs"):
                assert values == ["foo,bar", "baz"]
            elif key == "crs":
                assert values == ["EPSG:4326"]
            else:
                assert False, "unexpected key " + key

    def test_multiple_sets(self):
        nc_dict = NoCaseMultiDict()
        nc_dict["foo"] = "bar"
        assert nc_dict["FOO"] == "bar"
        nc_dict["foo"] = "baz"
        assert nc_dict["FOO"] == "baz"

    def test_missing_key(self):
        nc_dict = NoCaseMultiDict([("foo", "bar")])
        try:
            nc_dict["bar"]
            assert False, "Did not throw KeyError exception."
        except KeyError:
            pass

    def test_get(self):
        nc_dict = NoCaseMultiDict([("foo", "bar"), ("num", "42")])
        assert nc_dict.get("bar") is None
        assert nc_dict.get("bar", "default_bar") == "default_bar"
        assert nc_dict.get("num") == "42"
        assert nc_dict.get("num", type_func=int) == 42
        assert nc_dict.get("foo") == "bar"

    def test_get_all(self):
        nc_dict = NoCaseMultiDict([("foo", "bar"), ("num", "42"), ("foo", "biz")])
        assert nc_dict.get_all("bar") == []
        assert nc_dict.get_all("foo") == ["bar", "biz"]
        assert nc_dict.get_all("num") == ["42"]

    def test_set(self):
        nc_dict = NoCaseMultiDict()
        nc_dict.set("foo", "bar")
        assert nc_dict.get_all("fOO") == ["bar"]
        nc_dict.set("fOo", "buzz", append=True)
        assert nc_dict.get_all("FOO") == ["bar", "buzz"]
        nc_dict.set("foO", "bizz")
        assert nc_dict.get_all("FOO") == ["bizz"]
        nc_dict.set("foO", ["ham", "spam"], unpack=True)
        assert nc_dict.get_all("FOO") == ["ham", "spam"]
        nc_dict.set("FoO", ["egg", "bacon"], append=True, unpack=True)
        assert nc_dict.get_all("FOo") == ["ham", "spam", "egg", "bacon"]

    def test_setitem(self):
        nc_dict = NoCaseMultiDict()
        nc_dict["foo"] = "bar"
        assert nc_dict["foo"] == "bar"
        nc_dict["foo"] = "buz"
        assert nc_dict["foo"] == "buz"
        nc_dict["bar"] = nc_dict["foo"]
        assert nc_dict["bar"] == "buz"

        nc_dict["bing"] = "1"
        nc_dict["bong"] = "2"
        nc_dict["bing"] = nc_dict["bong"]
        assert nc_dict["bing"] == "2"
        assert nc_dict["bong"] == "2"

    def test_del(self):
        nc_dict = NoCaseMultiDict([("foo", "bar"), ("num", "42")])
        assert nc_dict["fOO"] == "bar"
        del nc_dict["FOO"]
        assert nc_dict.get("foo") is None


class DummyRequest(object):

    def __init__(self, args, url=""):
        self.args = args
        self.base_url = url


class TestWMSMapRequest(object):

    def setup_method(self):
        self.base_req = url_decode(
            """SERVICE=WMS&format=image%2Fpng&layers=foo&styles=&
REQUEST=GetMap&height=300&srs=EPSG%3A4326&VERSION=1.1.1&
bbox=7,50,8,51&width=400""".replace(
                "\n", ""
            )
        )


class TestWMS100MapRequest(TestWMSMapRequest):

    def setup_method(self):
        TestWMSMapRequest.setup_method(self)
        del self.base_req["service"]
        del self.base_req["version"]
        self.base_req["wmtver"] = "1.0.0"
        self.base_req["request"] = "Map"

    def test_basic_request(self):
        req = wms_request(DummyRequest(self.base_req), validate=False)
        assert isinstance(req, WMS100MapRequest)
        assert req.params.request == "GetMap"


class TestWMS111MapRequest(TestWMSMapRequest):

    def test_basic_request(self):
        req = wms_request(DummyRequest(self.base_req), validate=False)
        assert isinstance(req, WMS111MapRequest)
        assert req.params.request == "GetMap"


class TestWMS130MapRequest(TestWMSMapRequest):

    def setup_method(self):
        TestWMSMapRequest.setup_method(self)
        self.base_req["version"] = "1.3.0"
        self.base_req["crs"] = self.base_req["srs"]
        del self.base_req["srs"]

    def test_basic_request(self):
        req = wms_request(DummyRequest(self.base_req), validate=False)
        assert isinstance(req, WMS130MapRequest)
        assert req.params.request == "GetMap"
        assert req.params.bbox == (50.0, 7.0, 51.0, 8.0)

    def test_copy_with_request_params(self):
        # check that we allways have our internal axis order
        req1 = WMS130MapRequest(param=dict(bbox="10,0,20,40", crs="EPSG:4326"))
        assert req1.params.bbox == (0.0, 10.0, 40.0, 20.0)
        req2 = WMS111MapRequest(param=dict(bbox="0,10,40,20", srs="EPSG:4326"))
        assert req2.params.bbox == (0.0, 10.0, 40.0, 20.0)

        # 130 <- 111
        req3 = req1.copy_with_request_params(req2)
        assert req3.params.bbox == (0.0, 10.0, 40.0, 20.0)
        assert isinstance(req3, WMS130MapRequest)

        # 130 <- 130
        req4 = req1.copy_with_request_params(req3)
        assert req4.params.bbox == (0.0, 10.0, 40.0, 20.0)
        assert isinstance(req4, WMS130MapRequest)

        # 111 <- 130
        req5 = req2.copy_with_request_params(req3)
        assert req5.params.bbox == (0.0, 10.0, 40.0, 20.0)
        assert isinstance(req5, WMS111MapRequest)


class TestWMS111FeatureInfoRequest(TestWMSMapRequest):

    def setup_method(self):
        TestWMSMapRequest.setup_method(self)
        self.base_req["request"] = "GetFeatureInfo"
        self.base_req["x"] = "100"
        self.base_req["y"] = "150"
        self.base_req["query_layers"] = "foo"

    def test_basic_request(self):
        req = wms_request(DummyRequest(self.base_req))  # , validate=False)
        assert isinstance(req, WMS111FeatureInfoRequest)

    def test_pos(self):
        req = wms_request(DummyRequest(self.base_req))
        assert req.params.pos == (100, 150)

    def test_pos_coords(self):
        req = wms_request(DummyRequest(self.base_req))
        assert req.params.pos_coords == (7.25, 50.5)


class TestArcGISRequest(object):

    def test_base_request(self):
        req = ArcGISRequest(url="http://example.com/ArcGIS/rest/MapServer/")
        assert "http://example.com/ArcGIS/rest/MapServer/export" == req.url
        req.params.bbox = [-180.0, -90.0, 180.0, 90.0]
        assert (-180.0, -90.0, 180.0, 90.0) == req.params.bbox
        assert "-180.0,-90.0,180.0,90.0" == req.params["bbox"]
        req.params.size = [256, 256]
        assert (256, 256) == req.params.size
        assert "256,256" == req.params["size"]
        req.params.imageSR = "EPSG:4326"
        assert "4326" == req.params.imageSR
        assert "4326" == req.params["imageSR"]
        req.params.bboxSR = SRS("EPSG:4326")
        assert "4326" == req.params.bboxSR
        assert "4326" == req.params["bboxSR"]

    @pytest.mark.parametrize(
        "url,expected",
        [
            [
                "http://example.com/ArcGIS/rest/MapServer/",
                "http://example.com/ArcGIS/rest/MapServer/export",
            ],
            [
                "http://example.com/ArcGIS/rest/MapServer",
                "http://example.com/ArcGIS/rest/MapServer/export",
            ],
            [
                "http://example.com/ArcGIS/rest/MapServer/export",
                "http://example.com/ArcGIS/rest/MapServer/export",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer/",
                "http://example.com/ArcGIS/rest/ImageServer/exportImage",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer",
                "http://example.com/ArcGIS/rest/ImageServer/exportImage",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer/export",
                "http://example.com/ArcGIS/rest/ImageServer/exportImage",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer/exportImage",
                "http://example.com/ArcGIS/rest/ImageServer/exportImage",
            ],
            [
                "http://example.com/ArcGIS/rest/MapServer/export?param=foo",
                "http://example.com/ArcGIS/rest/MapServer/export?param=foo",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer/export?param=foo",
                "http://example.com/ArcGIS/rest/ImageServer/exportImage?param=foo",
            ],
        ],
    )
    def test_endpoint_urls(self, url, expected):
        req = ArcGISRequest(url=url)
        assert req.url == expected


class TestArcGISIndentifyRequest(object):

    def test_base_request(self):
        req = ArcGISIdentifyRequest(url="http://example.com/ArcGIS/rest/MapServer/")
        assert "http://example.com/ArcGIS/rest/MapServer/identify" == req.url
        req.params.bbox = [-180.0, -90.0, 180.0, 90.0]
        assert (-180.0, -90.0, 180.0, 90.0) == req.params.bbox
        assert "-180.0,-90.0,180.0,90.0" == req.params["mapExtent"]
        req.params.size = [256, 256]
        assert (256, 256) == req.params.size
        assert "256,256,96" == req.params["imageDisplay"]
        req.params.srs = "EPSG:4326"
        assert "EPSG:4326" == req.params.srs
        assert "4326" == req.params["sr"]

    @pytest.mark.parametrize(
        "url,expected",
        [
            [
                "http://example.com/ArcGIS/rest/MapServer/",
                "http://example.com/ArcGIS/rest/MapServer/identify",
            ],
            [
                "http://example.com/ArcGIS/rest/MapServer",
                "http://example.com/ArcGIS/rest/MapServer/identify",
            ],
            [
                "http://example.com/ArcGIS/rest/MapServer/export",
                "http://example.com/ArcGIS/rest/MapServer/identify",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer/",
                "http://example.com/ArcGIS/rest/ImageServer/identify",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer",
                "http://example.com/ArcGIS/rest/ImageServer/identify",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer/export",
                "http://example.com/ArcGIS/rest/ImageServer/identify",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer/exportImage",
                "http://example.com/ArcGIS/rest/ImageServer/identify",
            ],
            [
                "http://example.com/ArcGIS/rest/MapServer/export?param=foo",
                "http://example.com/ArcGIS/rest/MapServer/identify?param=foo",
            ],
            [
                "http://example.com/ArcGIS/rest/ImageServer/export?param=foo",
                "http://example.com/ArcGIS/rest/ImageServer/identify?param=foo",
            ],
        ],
    )
    def test_endpoint_urls(self, url, expected):
        req = ArcGISIdentifyRequest(url=url)
        assert req.url == expected


class TestRequest(object):

    def setup_method(self):
        self.env = {
            "HTTP_HOST": "localhost:5050",
            "PATH_INFO": "/service",
            "QUERY_STRING": "LAYERS=osm_mapnik&FORMAT=image%2Fpng&SPHERICALMERCATOR=true&SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&STYLES=&EXCEPTIONS=application%2Fvnd.ogc.se_inimage&SRS=EPSG%3A900913&bbox=1013566.9382067363,7051939.297837454,1030918.1436243634,7069577.142111099&WIDTH=908&HEIGHT=923",  # noqa
            "REMOTE_ADDR": "127.0.0.1",
            "REQUEST_METHOD": "GET",
            "SCRIPT_NAME": "",
            "SERVER_NAME": "127.0.0.1",
            "SERVER_PORT": "5050",
            "SERVER_PROTOCOL": "HTTP/1.1",
            "wsgi.url_scheme": "http",
        }

    def test_path(self):
        req = Request(self.env)
        assert req.path == "/service"

    def test_host_url(self):
        req = Request(self.env)
        assert req.host_url == "http://localhost:5050/"

    def test_base_url(self):
        req = Request(self.env)
        assert req.base_url == "http://localhost:5050/service"

        del self.env["HTTP_HOST"]
        req = Request(self.env)
        assert req.base_url == "http://127.0.0.1:5050/service"

        self.env["SERVER_PORT"] = "80"
        req = Request(self.env)
        assert req.base_url == "http://127.0.0.1/service"

    def test_query_string(self):
        self.env["QUERY_STRING"] = "Foo=boo&baz=baa&fOO=bizz"
        req = Request(self.env)
        print(req.args["foo"])
        assert req.args.get_all("foo") == ["boo", "bizz"]

    def test_query_string_encoding(self):
        env = {"QUERY_STRING": "foo=some%20special%20chars%20%26%20%3D"}
        req = Request(env)
        print(req.args["foo"])
        assert req.args["foo"] == u"some special chars & ="

    def test_script_url(self):
        req = Request(self.env)
        assert req.script_url == "http://localhost:5050"
        self.env["SCRIPT_NAME"] = "/"
        req = Request(self.env)
        assert req.script_url == "http://localhost:5050"

        self.env["SCRIPT_NAME"] = "/proxy"
        req = Request(self.env)
        assert req.script_url == "http://localhost:5050/proxy"

        self.env["SCRIPT_NAME"] = "/proxy/"
        req = Request(self.env)
        assert req.script_url == "http://localhost:5050/proxy"

    def test_pop_path(self):
        self.env["PATH_INFO"] = "/foo/service"
        req = Request(self.env)
        part = req.pop_path()
        assert part == "foo"
        assert self.env["PATH_INFO"] == "/service"
        assert self.env["SCRIPT_NAME"] == "/foo"

        part = req.pop_path()
        assert part == "service"
        assert self.env["PATH_INFO"] == ""
        assert self.env["SCRIPT_NAME"] == "/foo/service"

        part = req.pop_path()
        assert part == ""
        assert self.env["PATH_INFO"] == ""
        assert self.env["SCRIPT_NAME"] == "/foo/service"


def test_maprequest_from_request():
    env = {
        "QUERY_STRING": "layers=bar&bBOx=-90,-80,70.0,+80&format=image/png&"
        "WIdth=100&heIGHT=200&LAyerS=foo"
    }
    req = WMSMapRequest(param=Request(env).args)
    assert req.params.bbox == (-90.0, -80.0, 70.0, 80.0)
    assert req.params.layers == ["bar", "foo"]
    assert req.params.size == (100, 200)


class TestWMSMapRequestParams(object):

    def setup_method(self):
        self.m = WMSMapRequestParams(
            url_decode(
                "layers=bar&bBOx=-90,-80,70.0, 80&format=image/png"
                "&WIdth=100&heIGHT=200&LAyerS=foo&srs=EPSG%3A0815"
            )
        )

    def test_empty(self):
        m = WMSMapRequestParams()
        assert m.query_string == ""

    def test_size(self):
        assert self.m.size == (100, 200)
        self.m.size = (250, 350)
        assert self.m.size == (250, 350)
        assert self.m["width"] == "250"
        assert self.m["height"] == "350"
        del self.m["width"]
        assert self.m.size is None

    def test_format(self):
        assert self.m.format == "png"
        assert self.m.format_mime_type == "image/png"
        self.m["transparent"] = "True"
        assert self.m.format == "png"

    def test_bbox(self):
        assert self.m.bbox == (-90.0, -80.0, 70.0, 80.0)
        del self.m["bbox"]
        assert self.m.bbox is None
        self.m.bbox = (-90.0, -80.0, 70.0, 80.0)
        assert self.m.bbox == (-90.0, -80.0, 70.0, 80.0)
        self.m.bbox = "0.0, -40.0, 70.0, 80.0"
        assert self.m.bbox == (0.0, -40.0, 70.0, 80.0)
        self.m.bbox = None
        assert self.m.bbox is None

    def test_transparent(self):
        assert self.m.transparent is False
        self.m["transparent"] = "trUe"
        assert self.m.transparent is True

    def test_transparent_bool(self):
        self.m["transparent"] = True
        assert self.m["transparent"] == "True"

    def test_bgcolor(self):
        assert self.m.bgcolor == "#ffffff"
        self.m["bgcolor"] = "0x42cafe"
        assert self.m.bgcolor == "#42cafe"

    def test_srs(self):
        print(self.m.srs)
        assert self.m.srs == "EPSG:0815"
        del self.m["srs"]
        assert self.m.srs is None
        self.m.srs = SRS("EPSG:4326")
        assert self.m.srs == "EPSG:4326"

    def test_layers(self):
        assert list(self.m.layers) == ["bar", "foo"]

    def test_query_string(self):
        assert_query_eq(
            self.m.query_string,
            "layers=bar,foo&WIdth=100&bBOx=-90,-80,70.0,+80"
            "&format=image%2Fpng&srs=EPSG%3A0815&heIGHT=200",
        )

    def test_query_string_encoding(self):
        m = WMSMapRequestParams()
        m.layers = ["layer with whitespace", u"layer with ümlauts"]
        assert (
            m.query_string
            == "layers=layer%20with%20whitespace,layer%20with%20%C3%BCmlauts"
        )

    def test_get(self):
        assert self.m.get("LAYERS") == "bar"
        assert self.m.get("width", type_func=int) == 100

    def test_set(self):
        self.m.set("Layers", "baz", append=True)
        assert self.m.get("LAYERS") == "bar"
        self.m.set("Layers", "baz")
        assert self.m.get("LAYERS") == "baz"

    def test_attr_access(self):
        assert self.m["width"] == "100"
        assert self.m["height"] == "200"
        try:
            self.m.invalid
        except AttributeError:
            pass
        else:
            assert False

    def test_with_defaults(self):
        orig_req = WMSMapRequestParams(param=dict(layers="baz"))
        new_req = self.m.with_defaults(orig_req)
        assert new_req is not self.m
        assert self.m.get("LayErs") == "bar"
        assert new_req.get("LAyers") == "baz"
        assert new_req.size == (100, 200)


class TestURLDecode(object):

    def test_key_decode(self):
        d = url_decode("white+space=in+key&foo=bar", decode_keys=True)
        assert d["white space"] == "in key"
        assert d["foo"] == "bar"

    def test_include_empty(self):
        d = url_decode("bar&foo=baz&bing", include_empty=True)
        assert d["bar"] == ""
        assert d["foo"] == "baz"
        assert d["bing"] == ""


def test_non_mime_format():
    m = WMSMapRequest(param={"format": "jpeg"})
    assert m.params.format == "jpeg"


def test_request_w_url():
    url = WMSMapRequest(
        url="http://localhost:8000/service?", param={"layers": "foo,bar"}
    ).complete_url
    assert_url_eq(
        url,
        "http://localhost:8000/service?layers=foo,bar&styles=&request=GetMap&service=WMS",
    )
    url = WMSMapRequest(
        url="http://localhost:8000/service", param={"layers": "foo,bar"}
    ).complete_url
    assert_url_eq(
        url,
        "http://localhost:8000/service?layers=foo,bar&styles=&request=GetMap&service=WMS",
    )
    url = WMSMapRequest(
        url="http://localhost:8000/service?map=foo", param={"layers": "foo,bar"}
    ).complete_url
    assert_url_eq(
        url,
        "http://localhost:8000/service?map=foo&layers=foo,bar&styles=&request=GetMap&service=WMS",
    )


class TestWMSRequest(object):
    env = make_wsgi_env(
        """LAYERS=foo&FORMAT=image%2Fjpeg&SERVICE=WMS&VERSION=1.1.1&
REQUEST=GetMap&STYLES=&EXCEPTIONS=application%2Fvnd.ogc.se_xml&SRS=EPSG%3A900913&
BBOX=8,4,9,5&WIDTH=984&HEIGHT=708""".replace(
            "\n", ""
        )
    )

    def setup_method(self):
        self.req = Request(self.env)

    def test_valid_request(self):
        map_req = wms_request(self.req)
        # constructor validates
        assert map_req.params.size == (984, 708)

    def test_invalid_request(self):
        del self.req.args["request"]
        try:
            wms_request(self.req)
        except RequestError as e:
            assert "request" in e.msg
        else:
            assert False, "RequestError expected"

    def test_exception_handler(self):
        map_req = wms_request(self.req)
        assert isinstance(map_req.exception_handler, WMS111ExceptionHandler)

    def test_image_exception_handler(self):
        self.req.args["exceptions"] = "application/vnd.ogc.se_inimage"
        map_req = wms_request(self.req)
        assert isinstance(map_req.exception_handler, WMSImageExceptionHandler)

    def test_blank_exception_handler(self):
        self.req.args["exceptions"] = "blank"
        map_req = wms_request(self.req)
        assert isinstance(map_req.exception_handler, WMSBlankExceptionHandler)


class TestSRSAxisOrder(object):

    def setup_method(self):
        params111 = url_decode(
            """LAYERS=foo&FORMAT=image%2Fjpeg&SERVICE=WMS&
VERSION=1.1.1&REQUEST=GetMap&STYLES=&EXCEPTIONS=application%2Fvnd.ogc.se_xml&
SRS=EPSG%3A4326&BBOX=8,4,9,5&WIDTH=984&HEIGHT=708""".replace(
                "\n", ""
            )
        )
        self.req111 = WMS111MapRequest(params111)
        self.params130 = params111.copy()
        self.params130["version"] = "1.3.0"
        self.params130["crs"] = self.params130["srs"]
        del self.params130["srs"]

    def test_111_order(self):
        assert self.req111.params.bbox == (8, 4, 9, 5)

    def test_130_order_geog(self):
        req130 = WMS130MapRequest(self.params130)
        assert req130.params.bbox == (4, 8, 5, 9)
        self.params130["crs"] = "EPSG:4258"
        req130 = WMS130MapRequest(self.params130)
        assert req130.params.bbox == (4, 8, 5, 9)

    def test_130_order_geog_old(self):
        self.params130["crs"] = "CRS:84"
        req130 = WMS130MapRequest(self.params130)
        assert req130.params.bbox == (8, 4, 9, 5)

    def test_130_order_proj_north_east(self):
        self.params130["crs"] = "EPSG:31466"
        req130 = WMS130MapRequest(self.params130)
        assert req130.params.bbox == (4, 8, 5, 9)

    def test_130_order_proj(self):
        self.params130["crs"] = "EPSG:31463"
        req130 = WMS130MapRequest(self.params130)
        assert req130.params.bbox == (8, 4, 9, 5)


class TestTileRequest(object):

    def test_tms_request(self):
        env = {"PATH_INFO": "/tms/1.0.0/osm/5/2/3.png", "QUERY_STRING": ""}
        req = Request(env)
        tms = tile_request(req)
        assert isinstance(tms, TMSRequest)
        assert tms.tile == (2, 3, 5)
        assert tms.format == "png"
        assert tms.layer == "osm"
        assert tms.dimensions == {}

    def test_tile_request(self):
        env = {"PATH_INFO": "/tiles/1.0.0/osm/5/2/3.png", "QUERY_STRING": ""}
        req = Request(env)
        tile_req = tile_request(req)
        assert isinstance(tile_req, TileRequest)
        assert tile_req.tile == (2, 3, 5)
        assert tile_req.origin is None
        assert tile_req.format == "png"
        assert tile_req.layer == "osm"
        assert tile_req.dimensions == {}

    def test_tile_request_flipped_y(self):
        env = {"PATH_INFO": "/tiles/1.0.0/osm/5/2/3.png", "QUERY_STRING": "origin=nw"}
        req = Request(env)
        tile_req = tile_request(req)
        assert isinstance(tile_req, TileRequest)
        assert tile_req.tile == (2, 3, 5)  # not yet flipped
        assert tile_req.origin == "nw"
        assert tile_req.format == "png"
        assert tile_req.layer == "osm"
        assert tile_req.dimensions == {}

    def test_tile_request_w_epsg(self):
        env = {"PATH_INFO": "/tiles/1.0.0/osm/EPSG4326/5/2/3.png", "QUERY_STRING": ""}
        req = Request(env)
        tile_req = tile_request(req)
        assert isinstance(tile_req, TileRequest)
        assert tile_req.tile == (2, 3, 5)
        assert tile_req.format == "png"
        assert tile_req.layer == "osm"
        assert tile_req.dimensions == {"_layer_spec": "EPSG4326"}


def test_request_params_pickle():
    params = RequestParams(dict(foo="bar", zing="zong"))
    params2 = pickle.loads(pickle.dumps(params, 2))
    assert params.params == params2.params



# This file is part of the MapProxy project.
# Copyright (C) 2016 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from io import BytesIO

from mapproxy.request.wms import WMS111MapRequest
from mapproxy.test.image import is_png, create_tmp_image
from mapproxy.test.system import SysTest

import pytest


try:
    import boto3
    from moto import mock_aws
except ImportError:
    boto3 = None
    mock_aws = None


@pytest.fixture(scope="module")
def config_file():
    return "cache_s3.yaml"


@pytest.fixture(scope="module")
def s3_buckets():
    with mock_aws():
        boto3.client("s3").create_bucket(Bucket="default_bucket")
        boto3.client("s3").create_bucket(Bucket="tiles")
        boto3.client("s3").create_bucket(Bucket="reversetiles")

        yield


@pytest.mark.skipif(not (boto3 and mock_aws), reason="boto3 and moto required")
@pytest.mark.usefixtures("s3_buckets")
class TestS3Cache(SysTest):

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-150,-40,-140,-30",
                width="100",
                height="100",
                layers="default",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )

    def test_get_map_cached(self, app):
        # mock_aws interferes with MockServ, use boto to manually upload tile
        tile = create_tmp_image((256, 256))
        boto3.client("s3").upload_fileobj(
            BytesIO(tile),
            Bucket="default_bucket",
            Key="default_cache/WebMerc/4/1/9.png",
        )

        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)

    def test_get_map_cached_quadkey(self, app):
        # mock_aws interferes with MockServ, use boto to manually upload tile
        tile = create_tmp_image((256, 256))
        boto3.client("s3").upload_fileobj(
            BytesIO(tile), Bucket="tiles", Key="quadkeytiles/2003.png"
        )

        self.common_map_req.params.layers = "quadkey"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)

    def test_get_map_cached_reverse_tms(self, app):
        # mock_aws interferes with MockServ, use boto to manually upload tile
        tile = create_tmp_image((256, 256))
        boto3.client("s3").upload_fileobj(
            BytesIO(tile), Bucket="tiles", Key="reversetiles/9/1/4.png"
        )

        self.common_map_req.params.layers = "reverse"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)

    def test_get_map_cached_multiple_grids(self, app):
        # mock_aws interferes with MockServ, use boto to manually upload tile
        tile = create_tmp_image((256, 256))
        boto3.client("s3").upload_fileobj(
            BytesIO(tile), Bucket="tiles", Key="multiple_grids/WebMerc/9/1/4.png"
        )
        boto3.client("s3").upload_fileobj(
            BytesIO(tile), Bucket="tiles", Key="multiple_grids/WebMerc2/1/1/1.png"
        )

        resp = app.get("/wmts/multiple_grids/WebMerc/9/1/4.png")
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        resp = app.get("/wmts/multiple_grids/WebMerc2/1/1/1.png")
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)



# This file is part of the MapProxy project.
# Copyright (C) 2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from mapproxy.layer import MapQuery, InfoQuery
from mapproxy.srs import SRS
from mapproxy.service.wms import combined_layers
from mapproxy.source.wms import WMSSource
from mapproxy.client.wms import WMSClient
from mapproxy.request.wms import create_request


class TestCombinedLayers(object):
    q = MapQuery((0, 0, 10000, 10000), (100, 100), SRS(3857))

    def test_empty(self):
        assert combined_layers([], self.q) == []

    def test_same_source(self):
        layers = [
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "a"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "b"}, {}))
            ),
        ]
        combined = combined_layers(layers, self.q)
        assert len(combined) == 1
        assert combined[0].client.request_template.params.layers == ["a", "b"]

    def test_mixed_hosts(self):
        layers = [
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "a"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "b"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://bar/", "layers": "c"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://bar/", "layers": "d"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "e"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "f"}, {}))
            ),
        ]
        combined = combined_layers(layers, self.q)
        assert len(combined) == 3
        assert combined[0].client.request_template.params.layers == ["a", "b"]
        assert combined[1].client.request_template.params.layers == ["c", "d"]
        assert combined[2].client.request_template.params.layers == ["e", "f"]

    def test_mixed_params(self):
        layers = [
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "a"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "b"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "c"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "d"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "e"}, {}))
            ),
            WMSSource(
                WMSClient(create_request({"url": "http://foo/", "layers": "f"}, {}))
            ),
        ]

        layers[0].supported_srs = ["EPSG:4326"]
        layers[1].supported_srs = ["EPSG:4326"]

        layers[2].supported_formats = ["image/png"]
        layers[3].supported_formats = ["image/png"]

        combined = combined_layers(layers, self.q)
        assert len(combined) == 3
        assert combined[0].client.request_template.params.layers == ["a", "b"]
        assert combined[1].client.request_template.params.layers == ["c", "d"]
        assert combined[2].client.request_template.params.layers == ["e", "f"]


class TestInfoQuery(object):

    def test_coord(self):
        query = InfoQuery(
            (8, 50, 9, 51), (400, 1000), SRS(4326), (100, 600), "text/plain"
        )
        assert query.coord == (8.25, 50.4)



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import pytest

from mapproxy.test.system import SysTest


class TestResponseHeaders(SysTest):
    """
    Check if the vary header is set for text / xml content like capabilities
    """
    @pytest.fixture(scope='class')
    def config_file(self):
        return 'auth.yaml'

    def test_tms(self, app):
        resp = app.get('http://localhost/tms')
        assert resp.vary == ('X-Script-Name', 'X-Forwarded-Host', 'X-Forwarded-Proto')

    def test_wms(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities'
                       '&VERSION=1.1.0')
        assert resp.vary == ('X-Script-Name', 'X-Forwarded-Host', 'X-Forwarded-Proto')

    def test_wmts(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMTS&REQUEST=GetCapabilities')
        assert resp.vary == ('X-Script-Name', 'X-Forwarded-Host', 'X-Forwarded-Proto')

    def test_restful_wmts(self, app):
        resp = app.get('http://localhost/wmts/1.0.0/WMTSCapabilities.xml')
        assert resp.vary == ('X-Script-Name', 'X-Forwarded-Host', 'X-Forwarded-Proto')

    def test_no_endpoint(self, app):
        resp = app.get('http://localhost/service?', expect_errors=True)
        assert resp.vary == ('X-Script-Name', 'X-Forwarded-Host', 'X-Forwarded-Proto')

    def test_image_response(self, app):
        resp = app.get('http://localhost/tms/1.0.0/layer1a/EPSG900913/0/0/0.png')
        assert resp.vary is None



# -:- encoding: utf-8 -:-
# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy


class ImageOptions(object):
    def __init__(self, mode=None, transparent=None, opacity=None, resampling=None,
                 format=None, bgcolor=None, colors=None, encoding_options=None):
        self.transparent = transparent
        self.opacity = opacity
        self.resampling = resampling
        if format is not None:
            format = ImageFormat(format)
        self.format = format
        self.mode = mode
        self.bgcolor = bgcolor
        self.colors = colors
        self.encoding_options = encoding_options or {}

    def __repr__(self):
        options = []
        for k in dir(self):
            if k.startswith('_'):
                continue
            v = getattr(self, k)
            if v is not None and not hasattr(v, 'im_func') and not hasattr(v, '__func__'):
                options.append('%s=%r' % (k, v))
        return 'ImageOptions(%s)' % (', '.join(options), )

    def __eq__(self, other):
        if not isinstance(other, ImageOptions):
            return NotImplemented

        return (
            self.transparent == other.transparent
            and self.opacity == other.opacity
            and self.resampling == other.resampling
            and self.format == other.format
            and self.mode == other.mode
            and self.bgcolor == other.bgcolor
            and self.colors == other.colors
            and self.encoding_options == other.encoding_options
        )

    def copy(self):
        return copy.copy(self)


class ImageFormat(str):
    def __new__(cls, value, *args, **keywargs):
        if isinstance(value, ImageFormat):
            return value
        return str.__new__(cls, value)

    @property
    def mime_type(self):
        if self.startswith('image/'):
            return self
        return 'image/' + self

    @property
    def ext(self):
        ext = self
        if '/' in ext:
            ext = ext.split('/', 1)[1]
        if ';' in ext:
            ext = ext.split(';', 1)[0]

        return ext.strip()

    def __eq__(self, other):
        if isinstance(other, str):
            other = ImageFormat(other)
        elif not isinstance(other, ImageFormat):
            return NotImplemented

        return self.ext == other.ext

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not (self == other)


def create_image(size, image_opts=None):
    """
    Create a new image that is compatible with the given `image_opts`.
    Takes into account mode, transparent, bgcolor.
    """
    from PIL import Image, ImageColor

    if image_opts is None:
        mode = 'RGB'
        bgcolor = (255, 255, 255)
    else:
        mode = image_opts.mode
        if mode in (None, 'P'):
            if image_opts.transparent:
                mode = 'RGBA'
            else:
                mode = 'RGB'

        bgcolor = image_opts.bgcolor or (255, 255, 255)

        if isinstance(bgcolor, str):
            bgcolor = ImageColor.getrgb(bgcolor)

        if image_opts.transparent and len(bgcolor) == 3:
            bgcolor = bgcolor + (0, )

        if image_opts.mode == 'I':
            bgcolor = bgcolor[0]

    return Image.new(mode, size, bgcolor)


class ImageFormats(object):
    def __init__(self):
        self.format_options = {}

    def add(self, opts):
        assert opts.format is not None
        self.format_options[opts.format] = opts

    def options(self, format):
        opts = self.format_options.get(format)
        if not opts:
            opts = ImageOptions(transparent=False, format=format)
        return opts


def compatible_image_options(img_opts, base_opts=None):
    """
    Return ImageOptions that is compatible with all given `img_opts`.

    """
    if any(True for o in img_opts if o.colors == 0):
        colors = 0
    else:
        colors = max(o.colors or 0 for o in img_opts)

    transparent = None
    for o in img_opts:
        if o.transparent is False:
            transparent = False
            break
        if o.transparent is True:
            transparent = True

    if any(True for o in img_opts if o.mode):
        # I < P < RGB < RGBA :)
        mode = max(o.mode for o in img_opts if o.mode)
    else:
        mode = None

    if base_opts:
        options = base_opts.copy()
        if options.colors is None:
            options.colors = colors
        if options.mode is None:
            options.mode = mode
        if options.transparent is None:
            options.transparent = transparent
    else:
        options = img_opts[0].copy()
        options.colors = colors
        options.transparent = transparent
        options.mode = mode

    return options



# This file is part of the MapProxy project.
# Copyright (C) 2017 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import glob
import optparse
import os.path
import re
import sys
from collections import OrderedDict

from mapproxy.cache.compact import CompactCacheV1, CompactCacheV2
from mapproxy.cache.tile import Tile
from mapproxy.config import local_base_config
from mapproxy.config.loader import load_configuration, ConfigurationError

import logging
log = logging.getLogger('mapproxy.defrag')


def defrag_command(args=None):
    parser = optparse.OptionParser("%prog defrag-compact [options] -f mapproxy_conf")
    parser.add_option("-f", "--mapproxy-conf", dest="mapproxy_conf",
                      help="MapProxy configuration.")

    parser.add_option(
        "--min-percent", type=float, default=10.0,
        help="Only defrag if fragmentation is larger (10 means at least 10% of the file does not have to be used)")

    parser.add_option(
        "--min-mb", type=float, default=1.0,
        help="Only defrag if fragmentation is larger (2 means at least 2MB the file does not have to be used)")

    parser.add_option("--dry-run", "-n", action="store_true",
                      help="Do not de-fragment, only print output")

    parser.add_option("--caches", dest="cache_names", metavar='cache1,cache2,...',
                      help="only defragment the named caches")

    from mapproxy.script.util import setup_logging
    import logging
    setup_logging(logging.INFO, format="[%(asctime)s] %(msg)s")

    if args:
        args = args[1:]  # remove script name

    (options, args) = parser.parse_args(args)
    if not options.mapproxy_conf:
        parser.print_help()
        sys.exit(1)

    try:
        proxy_configuration = load_configuration(options.mapproxy_conf)
    except IOError as e:
        print('ERROR: ', "%s: '%s'" % (e.strerror, e.filename), file=sys.stderr)
        sys.exit(2)
    except ConfigurationError as e:
        print(e, file=sys.stderr)
        print('ERROR: invalid configuration (see above)', file=sys.stderr)
        sys.exit(2)

    with local_base_config(proxy_configuration.base_config):
        available_caches = OrderedDict()
        for name, cache_conf in proxy_configuration.caches.items():
            for grid, extent, tile_mgr in cache_conf.caches():
                if isinstance(tile_mgr.cache, (CompactCacheV1, CompactCacheV2)):
                    available_caches.setdefault(name, []).append(tile_mgr.cache)

        if options.cache_names:
            defrag_caches = options.cache_names.split(',')
            missing = set(defrag_caches).difference(available_caches.keys())
            if missing:
                print('unknown caches: %s' % (', '.join(missing), ))
                print('available compact caches: %s' %
                      (', '.join(available_caches.keys()), ))
                sys.exit(1)
        else:
            defrag_caches = None

        for name, caches in available_caches.items():
            if defrag_caches and name not in defrag_caches:
                continue
            for cache in caches:
                logger = DefragLog(name)
                defrag_compact_cache(cache,
                                     min_percent=options.min_percent/100,
                                     min_bytes=options.min_mb*1024*1024,
                                     dry_run=options.dry_run,
                                     log_progress=logger,
                                     )


def bundle_offset(fname):
    """
    >>> bundle_offset("path/to/R0000C0000.bundle")
    (0, 0)
    >>> bundle_offset("path/to/R0380C1380.bundle")
    (4992, 896)
    """
    match = re.search(r'R([A-F0-9]{4,})C([A-F0-9]{4,}).bundle$', fname, re.IGNORECASE)
    if match:
        r = int(match.group(1), 16)
        c = int(match.group(2), 16)
        return c, r


class DefragLog(object):
    def __init__(self, cache_name):
        self.cache_name = cache_name

    def log(self, fname, fragmentation, fragmentation_bytes, num, total, defrag):
        msg = "%s: %3d/%d (%s) fragmentation is %.1f%% (%dkb)" % (
            self.cache_name, num, total, fname, fragmentation, fragmentation_bytes/1024
        )
        if defrag:
            msg += " - defragmenting"
        else:
            msg += " - skipping"
        log.info(msg)


def defrag_compact_cache(cache, min_percent=0.1, min_bytes=1024*1024, log_progress=None, dry_run=False):
    bundles = glob.glob(os.path.join(cache.cache_dir, 'L??', 'R????C????.bundle'))

    for i, bundle_file in enumerate(bundles):
        offset = bundle_offset(bundle_file)
        b = cache.bundle_class(bundle_file.rstrip('.bundle'), offset)
        size, file_size = b.size()

        defrag = 1 - float(size) / file_size
        defrag_bytes = file_size - size

        skip = False
        if defrag < min_percent or defrag_bytes < min_bytes:
            skip = True

        if log_progress:
            log_progress.log(
                fname=bundle_file,
                fragmentation=defrag * 100,
                fragmentation_bytes=defrag_bytes,
                num=i+1, total=len(bundles),
                defrag=not skip,
            )

        if skip or dry_run:
            continue

        tmp_bundle = os.path.join(cache.cache_dir, 'tmp_defrag')
        defb = cache.bundle_class(tmp_bundle, offset)
        stored_tiles = False

        for y in range(128):
            tiles = [Tile((x, y, 0)) for x in range(128)]
            b.load_tiles(tiles)
            tiles = [t for t in tiles if t.source]
            if tiles:
                stored_tiles = True
                defb.store_tiles(tiles)

        # remove first
        # - in case bundle is empty
        # - windows does not support rename to existing files
        if os.path.exists(bundle_file):
            os.remove(bundle_file)
        if os.path.exists(bundle_file[:-1] + 'x'):
            os.remove(bundle_file[:-1] + 'x')

        if stored_tiles:
            os.rename(tmp_bundle + '.bundle', bundle_file)
            if os.path.exists(tmp_bundle + '.bundlx'):
                os.rename(tmp_bundle + '.bundlx', bundle_file[:-1] + 'x')
            if os.path.exists(tmp_bundle + '.lck'):
                os.unlink(tmp_bundle + '.lck')



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image
from mapproxy.image import ImageSource
from mapproxy.image.merge import concat_legends
from mapproxy.test.image import is_png


class Test_Concat_Legends(object):
    def test_concatenation(self):
        legends = []
        img_1 = Image.new(mode='RGBA', size=(30, 10), color="red")
        img_2 = Image.new(mode='RGBA', size=(10, 10), color="black")
        img_3 = Image.new(mode='RGBA', size=(50, 80), color="blue")
        legends.append(ImageSource(img_1))
        legends.append(ImageSource(img_2))
        legends.append(ImageSource(img_3))
        source = concat_legends(legends)
        src_img = source.as_image()
        assert src_img.getpixel((0, 90)) == (255, 0, 0, 255)
        assert src_img.getpixel((0, 80)) == (0, 0, 0, 255)
        assert src_img.getpixel((0, 0)) == (0, 0, 255, 255)
        assert src_img.getpixel((49, 99)) == (255, 255, 255, 0)
        assert is_png(source.as_buffer())



# This file is part of the MapProxy project.
# Copyright (C) 2010, 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division

import sys
from collections import deque
from contextlib import contextmanager
from itertools import zip_longest

try:
    import Queue
except ImportError:
    import queue as Queue

from mapproxy.config import base_config
from mapproxy.grid.meta_grid import MetaGrid
from mapproxy.source import SourceError
from mapproxy.config import local_base_config
from mapproxy.util.lock import LockTimeout
from mapproxy.seed.util import format_seed_task
from mapproxy.seed.cachelock import DummyCacheLocker, CacheLockedError

from mapproxy.seed.util import (exp_backoff, limit_sub_bbox,
                                status_symbol, BackoffError)

import logging
log = logging.getLogger(__name__)

NONE = 0
CONTAINS = -1
INTERSECTS = 1

# Decide whether to use multiprocessing or threading. multiprocessing should be faster but
# it is not well supported on all platforms. Especially regarding lambdas and anonymous
# function/classes which are used in proj.py for example.
#
# Since Python 3.8, MacOS uses a non-forking start method for multiprocessing which
# inhibits similar restrictions to Windows.
if sys.platform == 'win32' or sys.platform == 'darwin':
    import threading
    proc_class = threading.Thread
    queue_class = Queue.Queue
else:
    import multiprocessing
    proc_class = multiprocessing.Process
    queue_class = multiprocessing.Queue


class TileWorkerPool(object):
    """
    Manages multiple TileWorker.
    """

    def __init__(self, task, worker_class, size=2, dry_run=False, progress_logger=None):
        self.tiles_queue = queue_class(size)
        self.task = task
        self.dry_run = dry_run
        self.procs = []
        self.progress_logger = progress_logger
        conf = base_config()
        for _ in range(size):
            worker = worker_class(self.task, self.tiles_queue, conf)
            worker.start()
            self.procs.append(worker)

    def process(self, tiles, progress):
        if not self.dry_run:
            while True:
                try:
                    self.tiles_queue.put(tiles, timeout=5)
                except Queue.Full:
                    alive = False
                    for proc in self.procs:
                        if proc.is_alive():
                            alive = True
                            break
                    if not alive:
                        log.warning('no workers left, stopping')
                        raise SeedInterrupted
                    continue
                else:
                    break

            if self.progress_logger:
                self.progress_logger.log_step(progress)

    def stop(self, force=False):
        """
        Stop seed workers by sending None-sentinel and joining the workers.

        :param force: Skip sending None-sentinel and join with a timeout.
                      For use when workers might be shutdown already by KeyboardInterrupt.
        """
        if not force:
            alives = 0
            for proc in self.procs:
                if proc.is_alive():
                    alives += 1

            while alives:
                # put None-sentinels to queue as long as we have workers alive
                try:
                    self.tiles_queue.put(None, timeout=1)
                    alives -= 1
                except Queue.Full:
                    alives = 0
                    for proc in self.procs:
                        if proc.is_alive():
                            alives += 1

        if force:
            timeout = 1.0
        else:
            timeout = None
        for proc in self.procs:
            proc.join(timeout)


class TileWorker(proc_class):
    def __init__(self, task, tiles_queue, conf):
        proc_class.__init__(self)
        proc_class.daemon = True
        self.task = task
        self.tile_mgr = task.tile_manager
        self.tiles_queue = tiles_queue
        self.conf = conf

    def run(self):
        with local_base_config(self.conf):
            try:
                self.work_loop()
            except KeyboardInterrupt:
                return
            except BackoffError:
                return


class TileSeedWorker(TileWorker):
    def work_loop(self):
        while True:
            tiles = self.tiles_queue.get()
            if tiles is None:
                return
            with self.tile_mgr.session():
                exp_backoff(self.tile_mgr.load_tile_coords, args=(tiles,),
                            max_repeat=100, max_backoff=600,
                            exceptions=(SourceError, IOError), ignore_exceptions=(LockTimeout, ))


class TileCleanupWorker(TileWorker):
    def work_loop(self):
        while True:
            tiles = self.tiles_queue.get()
            if tiles is None:
                return
            with self.tile_mgr.session():
                self.tile_mgr.remove_tile_coords(tiles)


class SeedProgress(object):
    def __init__(self, old_progress_identifier=None):
        self.progress = 0.0
        self.level_progress_percentages = [1.0]
        self.level_progresses = None
        self.level_progresses_level = 0
        self.progress_str_parts = []
        self.old_level_progresses = old_progress_identifier

    def step_forward(self, subtiles=1):
        self.progress += self.level_progress_percentages[-1] / subtiles

    @property
    def progress_str(self):
        return ''.join(self.progress_str_parts)

    @contextmanager
    def step_down(self, i, subtiles):
        if self.level_progresses is None:
            self.level_progresses = []
        self.level_progresses = self.level_progresses[:self.level_progresses_level]
        self.level_progresses.append((i, subtiles))
        self.level_progresses_level += 1
        self.progress_str_parts.append(status_symbol(i, subtiles))
        self.level_progress_percentages.append(self.level_progress_percentages[-1] / subtiles)

        yield

        self.level_progress_percentages.pop()
        self.progress_str_parts.pop()

        self.level_progresses_level -= 1
        if self.level_progresses_level == 0:
            self.level_progresses = []

    def already_processed(self):
        return self.can_skip(self.old_level_progresses, self.level_progresses)

    def current_progress_identifier(self):
        if self.already_processed() or self.level_progresses is None:
            return self.old_level_progresses
        return self.level_progresses[:]

    @staticmethod
    def can_skip(old_progress, current_progress):
        """
        Return True if the `current_progress` is behind the `old_progress` -
        when it isn't as far as the old progress.

        >>> SeedProgress.can_skip(None, [(0, 4)])
        False
        >>> SeedProgress.can_skip([], [(0, 4)])
        True
        >>> SeedProgress.can_skip([(0, 4)], None)
        False
        >>> SeedProgress.can_skip([(0, 4)], [(0, 4)])
        False
        >>> SeedProgress.can_skip([(1, 4)], [(0, 4)])
        True
        >>> SeedProgress.can_skip([(0, 4)], [(0, 4), (0, 4)])
        False

        >>> SeedProgress.can_skip([(0, 4), (0, 4), (2, 4)], [(0, 4), (0, 4)])
        False
        >>> SeedProgress.can_skip([(0, 4), (0, 4), (2, 4)], [(0, 4), (0, 4), (1, 4)])
        True
        >>> SeedProgress.can_skip([(0, 4), (0, 4), (2, 4)], [(0, 4), (0, 4), (2, 4)])
        False
        >>> SeedProgress.can_skip([(0, 4), (0, 4), (2, 4)], [(0, 4), (0, 4), (3, 4)])
        False
        >>> SeedProgress.can_skip([(0, 4), (0, 4), (2, 4)], [(0, 4), (1, 4)])
        False
        >>> SeedProgress.can_skip([(0, 4), (0, 4), (2, 4)], [(0, 4), (1, 4), (0, 4)])
        False
        """
        if current_progress is None:
            return False
        if old_progress is None:
            return False
        if old_progress == []:
            return True
        for old, current in zip_longest(old_progress, current_progress, fillvalue=None):
            if old is None:
                return False
            if current is None:
                return False
            if old < current:
                return False
            if old > current:
                return True
        return False

    def running(self):
        return True


class StopProcess(Exception):
    pass


class SeedInterrupted(Exception):
    pass


class TileWalker(object):
    """
    TileWalker traverses through all tiles in a tile grid and calls worker_pool.process
    for each (meta) tile. It traverses the tile grid (pyramid) depth-first.
    Intersection with coverages are checked before handling subtiles in the next level,
    allowing to determine if all subtiles should be seeded or skipped.
    """

    def __init__(self, task, worker_pool, handle_stale=False, handle_uncached=False, handle_all=False,
                 work_on_metatiles=True, skip_geoms_for_last_levels=0, progress_logger=None,
                 seed_progress=None):
        self.tile_mgr = task.tile_manager
        self.task = task
        self.worker_pool = worker_pool
        self.handle_all = handle_all
        self.handle_stale = handle_stale
        self.handle_uncached = handle_uncached
        self.work_on_metatiles = work_on_metatiles
        self.skip_geoms_for_last_levels = skip_geoms_for_last_levels
        self.progress_logger = progress_logger

        num_seed_levels = len(task.levels)
        if num_seed_levels >= 4:
            self.report_till_level = task.levels[num_seed_levels-2]
        else:
            self.report_till_level = task.levels[num_seed_levels-1]
        meta_size = self.tile_mgr.meta_grid.meta_size if self.tile_mgr.meta_grid else (1, 1)
        self.tiles_per_metatile = meta_size[0] * meta_size[1]
        self.grid = MetaGrid(self.tile_mgr.grid, meta_size=meta_size, meta_buffer=0)
        self.count = 0
        self.seed_progress = seed_progress or SeedProgress()

        # It is possible that we 'walk' through the same tile multiple times
        # when seeding irregular tile grids[0]. limit_sub_bbox prevents that we
        # recurse into the same area multiple times, but it is still possible
        # that a tile is processed multiple times. Locking prevents that a tile
        # is seeded multiple times, but it is possible that we count the same tile
        # multiple times (in dry-mode, or while the tile is in the process queue).

        # Tile counts can be off by 280% with sqrt2 grids.
        # We keep a small cache of already processed tiles to skip most duplicates.
        # A simple cache of 64 tile coordinates for each level already brings the
        # difference down to ~8%, which is good enough and faster than a more
        # sophisticated FIFO cache with O(1) lookup, or even caching all tiles.

        # [0] irregular tile grids: where one tile does not have exactly 4 subtiles
        # Typically when you use res_factor, or a custom res list.
        self.seeded_tiles = {x: deque(maxlen=64) for x in task.levels}

    def walk(self):
        assert self.handle_stale or self.handle_uncached
        bbox = self.task.coverage.extent.bbox_for(self.tile_mgr.grid.srs)
        if self.seed_progress.already_processed():
            # nothing to seed
            self.seed_progress.step_forward()
        else:
            try:
                self._walk(bbox, self.task.levels)
            except StopProcess:
                pass
        self.report_progress(self.task.levels[0], self.task.coverage.bbox)

    def _walk(self, cur_bbox, levels, current_level=0, all_subtiles=False):
        """
        :param cur_bbox: the bbox to seed in this call
        :param levels: list of levels to seed
        :param all_subtiles: seed all subtiles and do not check for
                             intersections with bbox/geom
        """
        bbox_, tiles, subtiles = self.grid.get_affected_level_tiles(cur_bbox, current_level)
        total_subtiles = tiles[0] * tiles[1]
        if len(levels) < self.skip_geoms_for_last_levels:
            # do not filter in last levels
            all_subtiles = True
        subtiles = self._filter_subtiles(subtiles, all_subtiles)

        if current_level in levels and current_level <= self.report_till_level:
            self.report_progress(current_level, cur_bbox)

        if not self.seed_progress.running():
            if current_level in levels:
                self.report_progress(current_level, cur_bbox)
            self.tile_mgr.cleanup()
            raise StopProcess()

        process = False
        if current_level in levels:
            levels = levels[1:]
            process = True

        for i, (subtile, sub_bbox, intersection) in enumerate(subtiles):
            if subtile is None:  # no intersection
                self.seed_progress.step_forward(total_subtiles)
                continue
            if levels:  # recurse to next level
                sub_bbox = limit_sub_bbox(cur_bbox, sub_bbox)
                if intersection == CONTAINS:
                    all_subtiles = True
                else:
                    all_subtiles = False

                with self.seed_progress.step_down(i, total_subtiles):
                    if self.seed_progress.already_processed():
                        self.seed_progress.step_forward()
                    else:
                        self._walk(sub_bbox, levels, current_level=current_level+1,
                                   all_subtiles=all_subtiles)

            if not process:
                continue

            # check if subtile was already processed. see comment in __init__
            if subtile in self.seeded_tiles[current_level]:
                if not levels:
                    self.seed_progress.step_forward(total_subtiles)
                continue
            self.seeded_tiles[current_level].appendleft(subtile)

            if not self.work_on_metatiles:
                # collect actual tiles
                handle_tiles = self.grid.tile_list(subtile)
            else:
                handle_tiles = [subtile]

            if self.handle_all:
                handle_tiles = [t for t in handle_tiles if
                                t is not None]
            elif self.handle_uncached:
                handle_tiles = [t for t in handle_tiles if
                                t is not None and
                                not self.tile_mgr.is_cached(t)]
            elif self.handle_stale:
                handle_tiles = [t for t in handle_tiles if
                                t is not None and
                                self.tile_mgr.is_stale(t)]
            if handle_tiles:
                self.count += 1
                self.worker_pool.process(handle_tiles, self.seed_progress)

            if not levels:
                self.seed_progress.step_forward(total_subtiles)

        if len(levels) >= 4:
            # call cleanup to close open caches
            # for connection based caches
            self.tile_mgr.cleanup()

    def report_progress(self, level, bbox):
        if self.progress_logger:
            self.progress_logger.log_progress(self.seed_progress, level, bbox,
                                              self.count * self.tiles_per_metatile)

    def _filter_subtiles(self, subtiles, all_subtiles):
        """
        Return an iterator with all sub tiles.
        Yields (None, None, None) for non-intersecting tiles,
        otherwise (subtile, subtile_bbox, intersection).
        """
        for subtile in subtiles:
            if subtile is None:
                yield None, None, None
            else:
                sub_bbox = self.grid.meta_tile(subtile).bbox
                if all_subtiles:
                    intersection = CONTAINS
                else:
                    intersection = self.task.intersects(sub_bbox)
                if intersection:
                    yield subtile, sub_bbox, intersection
                else:
                    yield None, None, None


class SeedTask(object):
    def __init__(self, md, tile_manager, levels, refresh_timestamp, refresh_all, coverage):
        self.md = md
        self.tile_manager = tile_manager
        self.grid = tile_manager.grid
        self.levels = levels
        self.refresh_timestamp = refresh_timestamp
        self.refresh_all = refresh_all
        self.coverage = coverage

    @property
    def id(self):
        return self.md['name'], self.md['cache_name'], self.md['grid_name'], tuple(self.levels)

    def intersects(self, bbox):
        if self.coverage.contains(bbox, self.grid.srs):
            return CONTAINS
        if self.coverage.intersects(bbox, self.grid.srs):
            return INTERSECTS
        return NONE


class CleanupTask(object):
    """
    :param coverage: area for the cleanup
    :param complete_extent: ``True`` if `coverage` equals the extent of the grid
    """

    def __init__(self, md, tile_manager, levels, remove_timestamp, remove_all, coverage, complete_extent=False):
        self.md = md
        self.tile_manager = tile_manager
        self.grid = tile_manager.grid
        self.levels = levels
        self.remove_timestamp = remove_timestamp
        self.remove_all = remove_all
        self.coverage = coverage
        self.complete_extent = complete_extent

    @property
    def id(self):
        return 'cleanup', self.md['name'], self.md['cache_name'], self.md['grid_name']

    def intersects(self, bbox):
        if self.coverage.contains(bbox, self.grid.srs):
            return CONTAINS
        if self.coverage.intersects(bbox, self.grid.srs):
            return INTERSECTS
        return NONE


def seed(tasks, concurrency=2, dry_run=False, skip_geoms_for_last_levels=0,
         progress_logger=None, cache_locker=None, skip_uncached=False):
    if cache_locker is None:
        cache_locker = DummyCacheLocker()

    active_tasks = tasks[::-1]
    while active_tasks:
        task = active_tasks[-1]
        print(format_seed_task(task))

        wait = len(active_tasks) == 1
        try:
            with cache_locker.lock(task.md['cache_name'], no_block=not wait):
                if progress_logger and progress_logger.progress_store:
                    progress_logger.current_task_id = task.id
                    start_progress = progress_logger.progress_store.get(task.id)
                else:
                    start_progress = None
                seed_progress = SeedProgress(old_progress_identifier=start_progress)
                seed_task(task, concurrency, dry_run, skip_geoms_for_last_levels, progress_logger,
                          seed_progress=seed_progress, skip_uncached=skip_uncached)
        except CacheLockedError:
            print('    ...cache is locked, skipping')
            active_tasks = [task] + active_tasks[:-1]
        else:
            active_tasks.pop()


def seed_task(task, concurrency=2, dry_run=False, skip_geoms_for_last_levels=0,
              progress_logger=None, seed_progress=None, skip_uncached=False):
    if task.coverage is False:
        return
    if task.refresh_timestamp is not None:
        task.tile_manager._expire_timestamp = task.refresh_timestamp
    task.tile_manager.minimize_meta_requests = False

    work_on_metatiles = True
    if task.tile_manager.rescale_tiles:
        work_on_metatiles = False

    tile_worker_pool = TileWorkerPool(task, TileSeedWorker, dry_run=dry_run,
                                      size=concurrency, progress_logger=progress_logger)
    # If the configuration requests to only refresh tiles which are already in cache,
    # tile walker parameters shall be adapted
    handle_stale = skip_uncached
    handle_uncached = not skip_uncached
    handle_all = task.refresh_all
    tile_walker = TileWalker(task, tile_worker_pool, handle_uncached=handle_uncached, handle_stale=handle_stale,
                             handle_all=handle_all, skip_geoms_for_last_levels=skip_geoms_for_last_levels,
                             progress_logger=progress_logger, seed_progress=seed_progress,
                             work_on_metatiles=work_on_metatiles,
                             )
    try:
        tile_walker.walk()
    except KeyboardInterrupt:
        tile_worker_pool.stop(force=True)
        raise
    finally:
        tile_worker_pool.stop()



__all__ = ['b', 'basestring_', 'is_unicode', 'coerce_text']


def b(s):
    if isinstance(s, str):
        return s.encode('latin1')
    return bytes(s)


basestring_ = (bytes, str)


def is_unicode(obj):
    return isinstance(obj, str)


def coerce_text(v):
    if not isinstance(v, basestring_):
        attr = '__str__'
        if hasattr(v, attr):
            return str(v)
        else:
            return bytes(v)
    return v
